#!/usr/bin/env python3

# TODO Add remaining xml methods for Protein class

from QUBEKit.engines import RDKit, Element
from QUBEKit.utils import constants
from QUBEKit.utils.exceptions import FileTypeError, TopologyMismatch

from collections import OrderedDict
from datetime import datetime
from itertools import groupby
import os
from pathlib import Path
import pickle
import re

import networkx as nx
import numpy as np

import xml.etree.ElementTree as ET
from xml.dom.minidom import parseString


class Atom:
    """
    Class to hold all of the "per atom" information.
    All atoms in Molecule will have an instance of this Atom class to describe their properties.
    """

    def __init__(self, atomic_number, atom_index, atom_name='', partial_charge=None, formal_charge=None):

        self.atomic_number = atomic_number
        # The actual atomic name as per periodic table e.g. C, F, Pb, etc
        self.atomic_mass = Element().mass(atomic_number)
        self.atomic_symbol = Element().name(atomic_number)
        # The QUBEKit assigned name derived from the atomic name and its index e.g. C1, F8, etc
        self.atom_name = atom_name
        self.atom_index = atom_index
        self.partial_charge = partial_charge
        self.formal_charge = formal_charge
        self.atom_type = None
        self.bonds = []

    def add_bond(self, bonded_index):
        """
        Add a bond to the atom, this will make sure the bond has not already been described
        :param bonded_index: The index of the atom bonded to self
        """

        if bonded_index not in self.bonds:
            self.bonds.append(bonded_index)

    def __repr__(self):
        return f'{self.__class__.__name__}({self.__dict__!r})'

    def __str__(self):
        """
        Prints the Atom class objects' names and values one after another with new lines between each.
        """

        return_str = ''
        for key, val in self.__dict__.items():
            # Return all objects as {atom object name} = {atom object value(s)}.
            return_str += f'\n{key} = {val}\n'

        return return_str


class DefaultsMixin:
    """
    This class holds all of the default configs from the config file.
    It's effectively a placeholder for all of the attributes which may
    be changed by editing the config file(s).

    It's a mixin because:
        * Normal multiple inheritance doesn't make sense in this context
        * Composition would be a bit messier and may require stuff like:
            mol = Ligand('methane.pdb', 'methane')
            mol.defaults.threads
            >> 2

            rather than the nice clean:
            mol.threads
            >> 2
        * Mixin is cleaner and clearer with respect to super() calls.
    """

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self.theory = 'wB97XD'
        self.basis = '6-311++G(d,p)'
        self.vib_scaling = 0.957
        self.threads = 2
        self.memory = 2
        self.convergence = 'GAU_TIGHT'
        self.iterations = 350
        self.bonds_engine = 'psi4'
        self.density_engine = 'g09'
        self.charges_engine = 'chargemol'
        self.ddec_version = 6
        self.geometric = True
        self.solvent = True

        self.dih_start = -165
        self.increment = 15
        self.dih_end = 180
        self.t_weight = 'infinity'
        self.opt_method = 'BFGS'
        self.refinement_method = 'SP'
        self.tor_limit = 20
        self.div_index = 0
        self.parameter_engine = 'xml'
        self.l_pen = 0.0
        self.mm_opt_method = 'openmm'
        self.relative_to_global = False

        self.excited_state = False
        self.excited_theory = 'TDA'
        self.nstates = 3
        self.excited_root = 1
        self.use_pseudo = False
        self.pseudo_potential_block = ""

        self.chargemol = '/home/<QUBEKit_user>/chargemol_09_26_2017'
        self.log = 999


class Molecule:
    """Base class for ligands and proteins."""

    def __init__(self, mol_input, name=None):
        """
        # Namings
        smiles                  str; SMILES string for the molecule e.g. 'c1cc[nH]c1'
        filename                str; Full filename e.g. methane.pdb
        qc_json                 json dict; QC json data. Only used if loading from qcarchive/portal
        name                    str; Molecule name e.g. 'methane'

        # Structure
        coords                  Dict of numpy arrays of the coords where the keys are the input type ('mm', 'qm', etc)
        topology                networkx Graph() object. Contains connection information for molecule
        angles                  List of tuples; Shows angles based on atom indices (from 0) e.g. (1, 2, 4), (1, 2, 5)
        dihedrals               Dictionary of dihedral tuples stored under their common core bond
                                e.g. {(1,2): [(3, 1, 2, 6), (3, 1, 2, 7)]}
        improper_torsions
        rotatable               List of dihedral core tuples [(1,2)]
        bond_lengths            Dictionary of bond lengths stored under the bond tuple
                                e.g. {(1, 3): 1.115341203992107} (angstroms)
        dih_phis                Dictionary of the dihedral angles measured in the molecule object stored under the
                                dihedral tuple e.g. {(3, 1, 2, 6): -70.3506776877}  (degrees)
        angle_values            Dictionary of the angle values measured in the molecule object stored under the
                                angle tuple e.g. {(2, 1, 3): 107.2268} (degrees)
        symm_hs
        qm_energy

        # XML Info
        xml_tree                An XML class object containing the force field values
        AtomTypes               dict of lists; basic non-symmetrised atoms types for each atom in the molecule
                                e.g. {0, ['C1', 'opls_800', 'C800'], 1: ['H1', 'opls_801', 'H801'], ... }
        Residues                List of residue names in the sequence they are found in the protein
        extra_sites
        qm_scans                Dictionary of central scanned bonds and there energies and structures

        Parameters
        -------------------
        This section has different units due to it interacting with OpenMM

        HarmonicBondForce       Dictionary of equilibrium distances and force constants stored under the bond tuple.
                                {(1, 2): [0.108, 405.65]} (nano meters, kJ/mol)
        HarmonicAngleForce      Dictionary of equilibrium angles and force constants stored under the angle tuple
                                e.g. {(2, 1, 3): [2.094395, 150.00]} (radians, kJ/mol)
        PeriodicTorsionForce    Dictionary of lists of the torsions values [periodicity, k, phase] stored under the
                                dihedral tuple with an improper tag only for improper torsions
                                e.g. {(3, 1, 2, 6): [[1, 0.6, 0], [2, 0, 3.141592653589793], ... Improper]}
        NonbondedForce          OrderedDict; L-J params. Keys are atom index, vals are [charge, sigma, epsilon]

        combination             str; Combination rules e.g. 'opls'
        sites                   OrderedDict of virtual site parameters
                                e.g.{0: [(top nos parent, a .b), (p1, p2, p3), charge]}

        # QUBEKit Internals
        state                   str; Describes the stage the analysis is in for pickling and unpickling
        config_file             str or path; the config file used for the execution
        restart                 bool; is the current execution starting from the beginning (False) or restarting (True)?
        """

        self.mol_input = mol_input
        self.name = name

        # The three possible input types; these are set in read_input()
        self.filename = None
        self.smiles = None
        self.qc_json = None

        self.rdkit_mol = None

        # Structure
        self.coords = {'qm': [], 'mm': [], 'input': [], 'temp': [], 'traj': []}
        self.topology = None
        self.angles = None
        self.dihedrals = None
        self.improper_torsions = None
        self.rotatable = None
        self.bond_lengths = None
        self.atoms = None
        self.dih_phis = None
        self.angle_values = None
        self.symm_hs = None
        self.qm_energy = None
        self.charge = 0
        self.multiplicity = 1
        self.qm_scans = None
        self.scan_order = None
        self.descriptors = None

        # XML Info
        self.xml_tree = None
        self.AtomTypes = None
        self.Residues = None
        self.extra_sites = None
        self.HarmonicBondForce = None
        self.HarmonicAngleForce = None
        self.PeriodicTorsionForce = None
        self.NonbondedForce = None
        self.bond_types = None
        self.angle_types = None

        self.combination = None
        self.sites = None

        # QUBEKit internals
        self.state = None
        self.config_file = 'master_config.ini'
        self.restart = False
        self.atom_symmetry_classes = None

        # Read mol_input and generate mol info from file, smiles string or qc_json.
        self.read_input()
        self.check_names_are_unique()

    def __repr__(self):
        return f'{self.__class__.__name__}({self.__dict__!r})'

    def __str__(self, trunc=False):
        """
        Prints the Molecule class objects' names and values one after another with new lines between each.
        Mostly just used for logging, debugging and displaying the results at the end of a run.
        If trunc is set to True:
            Check the items being printed:
                If they are empty or None -> skip over them
                If they're short (<120 chars) -> print them as normal
                Otherwise -> print a truncated version of them.
        If trunc is set to False:
            Just print everything (all key: value pairs) as is with a little extra spacing.
        """

        return_str = ''

        if trunc:
            for key, val in self.__dict__.items():

                # Just checking (if val) won't work as truth table is ambiguous for length > 1 arrays
                # I know this is gross, but it's the best of a bad situation.
                try:
                    bool(val)
                # Catch numpy array truth table error
                except ValueError:
                    continue

                # Ignore NoneTypes and empty lists / dicts etc unless type is int (charge = 0 for example)
                if val is not None and (val or isinstance(val, int)):
                    return_str += f'\n{key} = '

                    # if it's smaller than 120 chars: print it as is. Otherwise print a version cut off with "...".
                    if len(str(key) + str(val)) < 120:
                        # Print the repr() not the str(). This means generator expressions etc appear too.
                        return_str += repr(val)
                    else:
                        return_str += repr(val)[:121 - len(str(key))] + '...'

        else:
            for key, val in self.__dict__.items():
                # Return all objects as {ligand object name} = {ligand object value(s)} without any special formatting.
                return_str += f'\n{key} = {repr(val)}\n'

        return return_str

    def read_input(self):
        """
        The base input reader used upon instancing the class; it will decide which reader to use
        based on the file suffix, smiles string or qc_json.
        """

        # Dicts are used by qc_json
        if isinstance(self.mol_input, dict):
            self.qc_json = self.mol_input
            self.read_qc_json()

        # Check if a file that exists is passed
        elif Path(self.mol_input).exists():
            self.filename = Path(self.mol_input)
            self.name = self.filename.stem
            self.read_file(self.filename, input_type='input')

        # If it's a string, and doesn't contain '.' (not a file that doesn't exist) then it's a smiles string
        elif isinstance(self.mol_input, str) and '.' not in self.mol_input:
            self.smiles = self.mol_input
            self.rdkit_mol = RDKit().smiles_to_rdkit_mol(self.smiles, name=self.name)
            self.mol_from_rdkit(self.rdkit_mol, input_type='input')

        else:
            raise RuntimeError('Cannot parse input. A valid file type, smiles string or qc json must be provided.')

    def read_file(self, input_file, input_type):
        """
        A general file reader that should not be used to instantiate the class,
        first we choose which sub file reader to use then ensure that the topology matches that of the original molecule,
        before storing the new coordinates in the requested type.
        :param input_file: The name of the file to be read
        :param input_type: The type of coordinates it contains ie QM geometry/ MM geometry
        :return: An updated ligand object if the topology matches.
        """

        input_file = Path(input_file)
        try:
            self.rdkit_mol = RDKit().read_file(input_file)
            self.mol_from_rdkit(self.rdkit_mol, input_type=input_type)
        except AttributeError:
            if input_type == 'input':
                print('Could not create ligand from RDKit using default readers')
            if input_file.suffix == '.pdb':
                self.read_pdb(input_file, input_type=input_type)
            elif input_file.suffix == '.mol2':
                self.read_mol2(input_file, input_type=input_type)
            elif input_file.suffix == '.xyz':
                self.read_xyz(input_file, input_type=input_type)
            else:
                raise FileTypeError('Unsupported file type.')

    def check_names_are_unique(self):
        """
        To prevent problems occurring with some atoms perceived to be the same,
        check the atom names to ensure they are all unique.
        If some are the same, reset all atom names to be: f'{atomic_symbol}{index}'.
        This ensure they are all unique.
        """

        atom_names = [atom.atom_name for atom in self.atoms]
        # If some atom names aren't unique
        if len(set(atom_names)) < len(atom_names):
            # Change the atom name only; everything else is the same as it was.
            self.atoms = [Atom(atomic_number=atom.atomic_number,
                               atom_index=atom.atom_index,
                               atom_name=f'{atom.atomic_symbol}{i}',
                               partial_charge=atom.partial_charge,
                               formal_charge=atom.formal_charge) for i, atom in enumerate(self.atoms)]

    def mol_from_rdkit(self, rdkit_molecule, input_type='input'):
        """
        Unpack a RDKit molecule into the QUBEKit ligand if instance else just load a valid set of coordinates
        :param rdkit_molecule: The rdkit molecule instance
        :param input_type: Where the coordinates should be stored
        :return: The ligand object with the internal structures if instance
        """

        topology = nx.Graph()
        atoms = []

        if self.name is None:
            self.name = rdkit_molecule.GetProp('_Name')
        # Collect the atom names and bonds
        for atom in rdkit_molecule.GetAtoms():
            # Collect info about each atom
            try:
                # PDB file extraction
                atom_name = atom.GetMonomerInfo().GetName().strip()
            except AttributeError:
                try:
                    # Mol2 file extraction
                    atom_name = atom.GetProp('_TriposAtomName')
                except KeyError:
                    # This is ether from a smiles string or a mol file so pass for now and get the general info
                    pass

            atomic_number = atom.GetAtomicNum()
            index = atom.GetIdx()
            # smiles and mol files have no atom names so generate them here if they are not declared
            try:
                atom_name
            except NameError:
                atom_name = f'{atom.GetSymbol()}{index}'

            # Instance the basic qube_atom
            qube_atom = Atom(atomic_number, index, atom_name, formal_charge=atom.GetFormalCharge())
            qube_atom.atom_type = atom.GetSmarts()

            # Add the atoms as nodes
            topology.add_node(atom.GetIdx())

            # Add the bonds
            for bonded in atom.GetNeighbors():
                topology.add_edge(atom.GetIdx(), bonded.GetIdx())
                qube_atom.add_bond(bonded.GetIdx())

            # Now add the atom to the molecule
            atoms.append(qube_atom)

        # Now get the coordinates
        coords = rdkit_molecule.GetConformer().GetPositions()

        # Now get any descriptors we can find
        descriptors = RDKit().rdkit_descriptors(rdkit_molecule)

        self._validate_info(topology, atoms, coords, input_type, descriptors)

    def _validate_info(self, topology, atoms, coords, input_type, descriptors=None):
        """
        Check if the provided infomation should be stored or not
        :param topology: networkx graph of the topology
        :param atoms: a list of Atom objects
        :param coords: a numpy array of the coords
        :param descriptors: a dictionary of the rdkit descriptors
        :return: the updated ligand object
        """

        # Now check we instancing the ligand if we are then store the info
        if input_type == 'input':
            self.topology = topology
            self.atoms = atoms
            self.descriptors = descriptors
            self.coords[input_type] = coords
        else:
            # Check if the new topology is the same then store the new coordinates
            if nx.algorithms.is_isomorphic(self.topology, topology):
                self.coords[input_type] = coords
            else:
                raise TopologyMismatch('Topologies are not the same cannot store coordinates.')

    def read_pdb(self, input_file, input_type='input'):
        """
        Reads the input PDB file to find the ATOM or HETATM tags, extracts the elements and xyz coordinates.
        Then reads through the connection tags and builds a connectivity network
        (only works if connections are present in PDB file).
        Bonds are easily found through the edges of the network.
        Can also generate a simple plot of the network.
        """

        with open(input_file, 'r') as pdb:
            lines = pdb.readlines()

        molecule = []
        topology = nx.Graph()
        atoms = []

        # atom counter used for graph node generation
        atom_count = 0
        for line in lines:
            if 'ATOM' in line or 'HETATM' in line:
                # start collecting the atom class info
                atomic_symbol = str(line[76:78])
                atomic_symbol = re.sub('[0-9]+', '', atomic_symbol)
                atomic_symbol = atomic_symbol.strip()
                atom_name = str(line.split()[2])

                # If the element column is missing from the pdb, extract the atomic_symbol from the atom name.
                if not atomic_symbol:
                    atomic_symbol = str(line.split()[2])[:-1]
                    atomic_symbol = re.sub('[0-9]+', '', atomic_symbol)

                atomic_number = Element().number(atomic_symbol)
                # Now instance the qube atom
                qube_atom = Atom(atomic_number, atom_count, atom_name)
                atoms.append(qube_atom)

                # Also add the atom number as the node in the graph
                topology.add_node(atom_count)
                atom_count += 1
                molecule.append([float(line[30:38]), float(line[38:46]), float(line[46:54])])

            if 'CONECT' in line:
                atom_index = int(line.split()[1]) - 1
                # Now look through the connectivity section and add all edges to the graph corresponding to the bonds.
                for i in range(2, len(line.split())):
                    if int(line.split()[i]) != 0:
                        bonded_index = int(line.split()[i]) - 1
                        topology.add_edge(atom_index, bonded_index)
                        atoms[atom_index].add_bond(bonded_index)
                        atoms[bonded_index].add_bond(atom_index)

        # put the object back into the correct place
        coords = np.array(molecule)
        self._validate_info(topology, atoms, coords, input_type)

    def read_mol2(self, input_file, input_type='input'):
        """
        Read an input mol2 file and extract the atom names, positions, atom types and bonds.
        :param input_type: Assign the structure to right holder, input, mm, qm, temp or traj.
        :return: The object back into the right place.
        """

        molecule = []
        topology = nx.Graph()
        atoms = []

        # atom counter used for graph node generation
        atom_count = 0

        with open(input_file, 'r') as mol2:
            lines = mol2.readlines()

        atom_flag = False
        bond_flag = False

        for line in lines:
            if '@<TRIPOS>ATOM' in line:
                atom_flag = True
                continue
            elif '@<TRIPOS>BOND' in line:
                atom_flag = False
                bond_flag = True
                continue
            elif '@<TRIPOS>SUBSTRUCTURE' in line:
                bond_flag = False
                continue

            if atom_flag:
                # Add the molecule information
                atomic_symbol = line.split()[1][:2]
                atomic_symbol = re.sub('[0-9]+', '', atomic_symbol)
                atomic_symbol = atomic_symbol.strip().title()

                atomic_number = Element().number(atomic_symbol)

                molecule.append([float(line.split()[2]), float(line.split()[3]), float(line.split()[4])])

                # Collect the atom names
                atom_name = str(line.split()[1])

                # Add the nodes to the topology object
                topology.add_node(atom_count)
                atom_count += 1

                # Get the atom types
                atom_type = line.split()[5]
                atom_type = atom_type.replace(".", "")

                # Make the qube_atom
                qube_atom = Atom(atomic_number, atom_count, atom_name)
                qube_atom.atom_type = atom_type

                atoms.append(qube_atom)

            if bond_flag:
                # Add edges to the topology network
                atom_index, bonded_index = int(line.split()[1]) - 1, int(line.split()[2]) - 1
                topology.add_edge(atom_index, bonded_index)
                atoms[atom_index].add_bond(bonded_index)
                atoms[bonded_index].add_bond(atom_index)

        # put the object back into the correct place
        coords = np.array(molecule)
        self._validate_info(topology, atoms, coords, input_type)

    def read_xyz(self, name, input_type='traj'):
        """
        Read an xyz file and get all frames from the file and put in the traj molecule holder by default
        or if there is only one frame change the input location.
        """

        traj_molecules = []
        molecule = []
        try:
            with open(name, 'r') as xyz_file:
                # get the number of atoms
                n_atoms = len(self.atoms)
                for line in xyz_file:
                    line = line.split()
                    # skip frame heading lines
                    if len(line) <= 1:
                        next(xyz_file)
                        continue
                    molecule.append([float(line[1]), float(line[2]), float(line[3])])

                    if len(molecule) == n_atoms:
                        # we have collected the molecule now store the frame
                        traj_molecules.append(np.array(molecule))
                        molecule = []

            # check how many frames we have
            if len(traj_molecules) == 1:
                self.coords[input_type] = traj_molecules[0]
            else:
                self.coords[input_type] = traj_molecules

        except FileNotFoundError:
            raise FileNotFoundError('Cannot find xyz file to read.')

    def read_qc_json(self):
        """
        Using the QC json, extract the atoms and bonds (connectivity) to build a full topology.
        Insert the coords into the molecule too.
        """
        self.topology = nx.Graph()
        self.atoms = []

        for i, atom in enumerate(self.qc_json['symbols']):
            self.atoms.append(Atom(atomic_number=Element().number(atom), atom_index=i, atom_name=f'{atom}{i}'))
            self.topology.add_node(i)

        for bond in self.qc_json['connectivity']:
            self.topology.add_edge(*bond[:2])

        self.coords['input'] = np.array(self.qc_json['geometry']).reshape((len(self.atoms), 3)) * constants.BOHR_TO_ANGS

    def get_atom_with_name(self, name):
        """
        Search through the molecule for an atom with that name and return it when found
        :param name: The name of the atom we are looking for
        :return: The QUBE Atom object with the name
        """

        for atom in self.atoms:
            if atom.atom_name == name:
                return atom
        raise AttributeError('No atom found with that name.')

    def read_geometric_traj(self, trajectory):
        """
        Read in the molecule coordinates to the traj holder from a geometric optimisation using qcengine.
        :param trajectory: The qcengine trajectory
        """

        for frame in trajectory:
            opt_traj = []
            # Convert coordinates from bohr to angstroms
            geometry = np.array(frame['molecule']['geometry']) * constants.BOHR_TO_ANGS
            for i, atom in enumerate(frame['molecule']['symbols']):
                opt_traj.append([geometry[0 + i * 3], geometry[1 + i * 3], geometry[2 + i * 3]])
            self.coords['traj'].append(np.array(opt_traj))

    def find_impropers(self):
        """
        Take the topology graph and find all of the improper torsions in the molecule;
        these are atoms with 3 bonds.
        """

        improper_torsions = []

        for node in self.topology.nodes:
            near = sorted(list(nx.neighbors(self.topology, node)))
            # if the atom has 3 bonds it could be an improper
            # Check if an sp2 carbon
            if len(near) == 3 and self.atoms[node].atomic_symbol == 'C':
                improper_torsions.append((node, near[0], near[1], near[2]))

        if improper_torsions:
            self.improper_torsions = improper_torsions

    def find_angles(self):
        """
        Take the topology graph network and return a list of all angle combinations.
        Checked against OPLS-AA on molecules containing 10-63 angles.
        """

        angles = []

        for node in self.topology.nodes:
            bonded = sorted(list(nx.neighbors(self.topology, node)))

            # Check that the atom has more than one bond
            if len(bonded) < 2:
                continue

            # Find all possible angle combinations from the list
            for i in range(len(bonded)):
                for j in range(i + 1, len(bonded)):
                    atom1, atom3 = bonded[i], bonded[j]

                    angles.append((atom1, node, atom3))

        if angles:
            self.angles = angles

    def find_bond_lengths(self, input_type='input'):
        """For the given molecule and topology find the length of all of the bonds."""

        bond_lengths = {}

        molecule = self.coords[input_type]

        for edge in self.topology.edges:
            atom1 = molecule[edge[0]]
            atom2 = molecule[edge[1]]
            bond_lengths[edge] = np.linalg.norm(atom2 - atom1)

        # Check if the dictionary is full then store else leave as None
        if bond_lengths:
            self.bond_lengths = bond_lengths

    def find_dihedrals(self):
        """
        Take the topology graph network and again return a dictionary of all possible dihedral combinations
        stored under the central bond keys, which describe the angle.
        """

        dihedrals = {}

        # Work through the network using each edge as a central dihedral bond
        for edge in self.topology.edges:

            for start in list(nx.neighbors(self.topology, edge[0])):

                # Check atom not in main bond
                if start != edge[0] and start != edge[1]:

                    for end in list(nx.neighbors(self.topology, edge[1])):

                        # Check atom not in main bond
                        if end != edge[0] and end != edge[1]:

                            if edge not in dihedrals:
                                # Add the central edge as a key the first time it is used
                                dihedrals[edge] = [(start, edge[0], edge[1], end)]

                            else:
                                # Add the tuple to the correct key.
                                dihedrals[edge].append((start, edge[0], edge[1], end))

        if dihedrals:
            self.dihedrals = dihedrals

    def find_rotatable_dihedrals(self):
        """
        For each dihedral in the topology graph network and dihedrals dictionary, work out if the torsion is
        rotatable. Returns a list of dihedral dictionary keys representing the rotatable dihedrals.
        Also exclude standard rotations such as amides and methyl groups.
        """

        if self.dihedrals:
            rotatable = []

            # For each dihedral key remove the edge from the network
            for key in self.dihedrals:
                self.topology.remove_edge(*key)

                # Check if there is still a path between the two atoms in the edges.
                if not nx.has_path(self.topology, *key):
                    rotatable.append(key)

                # Add edge back to the network and try next key
                self.topology.add_edge(*key)

            if rotatable:
                self.rotatable = rotatable

    def get_dihedral_values(self, input_type='input'):
        """
        Taking the molecules' xyz coordinates and dihedrals dictionary, return a dictionary of dihedral
        angle keys and values. Also an option to only supply the keys of the dihedrals you want to calculate.
        """

        if self.dihedrals:

            dih_phis = {}

            molecule = self.coords[input_type]

            for val in self.dihedrals.values():
                for torsion in val:
                    # Calculate the dihedral angle in the molecule using the molecule data array.
                    x1, x2, x3, x4 = [molecule[torsion[i]] for i in range(4)]
                    b1, b2, b3 = x2 - x1, x3 - x2, x4 - x3
                    t1 = np.linalg.norm(b2) * np.dot(b1, np.cross(b2, b3))
                    t2 = np.dot(np.cross(b1, b2), np.cross(b2, b3))
                    dih_phis[torsion] = np.degrees(np.arctan2(t1, t2))

            if dih_phis:
                self.dih_phis = dih_phis

    def get_angle_values(self, input_type='input'):
        """
        For the given molecule and list of angle terms measure the angle values,
        then return a dictionary of angles and values.
        """

        angle_values = {}

        molecule = self.coords[input_type]

        for angle in self.angles:
            x1 = molecule[angle[0]]
            x2 = molecule[angle[1]]
            x3 = molecule[angle[2]]
            b1, b2 = x1 - x2, x3 - x2
            cosine_angle = np.dot(b1, b2) / (np.linalg.norm(b1) * np.linalg.norm(b2))
            angle_values[angle] = np.degrees(np.arccos(cosine_angle))

        if angle_values:
            self.angle_values = angle_values

    def symmetrise_bonded_parameters(self):
        """
        Apply symmetry to the parameters stored in the molecule based on types from rdkit.
        """

        if self.bond_types is not None:

            # Collect all of the bond values from the HarmonicBondForce dict
            for bonds in self.bond_types.values():
                bond_lens, bond_forces = zip(*[self.HarmonicBondForce[bond] for bond in bonds])

                # Average
                bond_lens, bond_forces = sum(bond_lens) / len(bond_lens), sum(bond_forces) / len(bond_forces)

                # Replace with averaged values
                for bond in bonds:
                    self.HarmonicBondForce[bond] = [bond_lens, bond_forces]

            # Collect all of the angle values from the HarmonicAngleForce dict
            for angles in self.angle_types.values():
                angle_vals, angle_forces = zip(*[self.HarmonicAngleForce[angle] for angle in angles])

                # Average
                angle_vals, angle_forces = sum(angle_vals) / len(angle_vals), sum(angle_forces) / len(angle_forces)

                # Replace with averaged values
                for angle in angles:
                    self.HarmonicAngleForce[angle] = [angle_vals, angle_forces]

    def write_parameters(self, name=None, is_protein=False):
        """Take the molecule's parameter set and write an xml file for the molecule."""

        # First build the xml tree
        self.build_tree(protein=is_protein)

        tree = self.xml_tree.getroot()
        messy = ET.tostring(tree, 'utf-8')

        pretty_xml_as_string = parseString(messy).toprettyxml(indent="")

        with open(f'{name if name is not None else self.name}.xml', 'w+') as xml_doc:
            xml_doc.write(pretty_xml_as_string)

    def build_tree(self, protein):
        """Separates the parameters and builds an xml tree ready to be used."""

        # Create XML layout
        root = ET.Element('ForceField')
        AtomTypes = ET.SubElement(root, "AtomTypes")
        Residues = ET.SubElement(root, "Residues")

        Residue = ET.SubElement(Residues, "Residue", name=f'{"QUP" if protein else "UNK"}')

        HarmonicBondForce = ET.SubElement(root, "HarmonicBondForce")
        HarmonicAngleForce = ET.SubElement(root, "HarmonicAngleForce")
        PeriodicTorsionForce = ET.SubElement(root, "PeriodicTorsionForce")

        # Assign the combination rule
        c14 = '0.83333' if self.combination == 'amber' else '0.5'
        l14 = '0.5'

        # add the combination rule to the xml for geometric.
        NonbondedForce = ET.SubElement(root, "NonbondedForce", attrib={
            'coulomb14scale': c14, 'lj14scale': l14,
            'combination': self.combination})

        for key, val in self.AtomTypes.items():
            ET.SubElement(AtomTypes, "Type", attrib={
                'name': val[1], 'class': val[2],
                'element': self.atoms[key].atomic_symbol,
                'mass': str(self.atoms[key].atomic_mass)})

            ET.SubElement(Residue, "Atom", attrib={'name': val[0], 'type': val[1]})

        # Add the bonds / connections
        for key, val in self.HarmonicBondForce.items():
            ET.SubElement(Residue, "Bond", attrib={'from': str(key[0]), 'to': str(key[1])})

            ET.SubElement(HarmonicBondForce, "Bond", attrib={
                'class1': self.AtomTypes[key[0]][2],
                'class2': self.AtomTypes[key[1]][2],
                'length': f'{val[0]:.6f}', 'k': f'{val[1]:.6f}'})

        # Add the angles
        for key, val in self.HarmonicAngleForce.items():
            ET.SubElement(HarmonicAngleForce, "Angle", attrib={
                'class1': self.AtomTypes[key[0]][2],
                'class2': self.AtomTypes[key[1]][2],
                'class3': self.AtomTypes[key[2]][2],
                'angle': f'{val[0]:.6f}', 'k': f'{val[1]:.6f}'})

        # add the proper and improper torsion terms
        for key in self.PeriodicTorsionForce:
            if self.PeriodicTorsionForce[key][-1] == 'Improper':
                tor_type = 'Improper'
            else:
                tor_type = 'Proper'
            ET.SubElement(PeriodicTorsionForce, tor_type, attrib={
                'class1': self.AtomTypes[key[0]][2],
                'class2': self.AtomTypes[key[1]][2],
                'class3': self.AtomTypes[key[2]][2],
                'class4': self.AtomTypes[key[3]][2],
                'k1': str(self.PeriodicTorsionForce[key][0][1]),
                'k2': str(self.PeriodicTorsionForce[key][1][1]),
                'k3': str(self.PeriodicTorsionForce[key][2][1]),
                'k4': str(self.PeriodicTorsionForce[key][3][1]),
                'periodicity1': '1', 'periodicity2': '2',
                'periodicity3': '3', 'periodicity4': '4',
                'phase1': str(self.PeriodicTorsionForce[key][0][2]),
                'phase2': str(self.PeriodicTorsionForce[key][1][2]),
                'phase3': str(self.PeriodicTorsionForce[key][2][2]),
                'phase4': str(self.PeriodicTorsionForce[key][3][2])})

        # add the non-bonded parameters
        for key in self.NonbondedForce:
            ET.SubElement(NonbondedForce, "Atom", attrib={
                'type': self.AtomTypes[key][1],
                'charge': f'{self.NonbondedForce[key][0]:.6f}',
                'sigma': f'{self.NonbondedForce[key][1]:.6f}',
                'epsilon': f'{self.NonbondedForce[key][2]:.6f}'})

        # Add all of the virtual site info if present
        if self.sites:
            # Add the atom type to the top
            for key, val in self.sites.items():
                ET.SubElement(AtomTypes, "Type", attrib={
                    'name': f'v-site{key + 1}', 'class': f'X{key + 1}', 'mass': '0'})

                # Add the atom info
                ET.SubElement(Residue, "Atom", attrib={
                    'name': f'X{key + 1}', 'type': f'v-site{key + 1}'})

                # Add the local coords site info
                ET.SubElement(Residue, "VirtualSite", attrib={
                    'type': 'localCoords',
                    'index': str(key + len(self.atoms)),
                    'atom1': str(val[0][0]), 'atom2': str(val[0][1]), 'atom3': str(val[0][2]),
                    'wo1': '1.0', 'wo2': '0.0', 'wo3': '0.0',
                    'wx1': '-1.0', 'wx2': '1.0', 'wx3': '0.0',
                    'wy1': '-1.0', 'wy2': '0.0', 'wy3': '1.0',
                    'p1': f'{float(val[1][0]):.4f}',
                    'p2': f'{float(val[1][1]):.4f}',
                    'p3': f'{float(val[1][2]):.4f}'})

                # Add the nonbonded info
                ET.SubElement(NonbondedForce, "Atom", attrib={
                    'type': f'v-site{key + 1}',
                    'charge': f'{val[2]}',
                    'sigma': '1.000000',
                    'epsilon': '0.000000'})

        # Store the tree back into the molecule
        self.xml_tree = ET.ElementTree(root)

    def write_xyz(self, input_type='input', name=None):
        """
        Write a general xyz file of the molecule if there are multiple geometries in the molecule write a traj
        :param input_type: Where the molecule coordinates are taken from
        :param name: The name of the xyz file to be produced; otherwise self.name is used.
        """

        with open(f'{name if name is not None else self.name}.xyz', 'w+') as xyz_file:

            if len(self.coords[input_type]) / len(self.atoms) == 1:
                message = 'xyz file generated with QUBEKit'
                end = ''
                trajectory = [self.coords[input_type]]

            else:
                message = f'QUBEKit xyz trajectory FRAME '
                end = 1
                trajectory = self.coords[input_type]

            # Write out each frame
            for frame in trajectory:

                xyz_file.write(f'{len(self.atoms)}\n')
                xyz_file.write(f'{message}{end}\n')

                for i, atom in enumerate(frame):
                    xyz_file.write(
                        f'{self.atoms[i].atomic_symbol}       {atom[0]: .10f}   {atom[1]: .10f}   {atom[2]: .10f}\n')

                try:
                    end += 1
                except TypeError:
                    # This is the result of only printing one frame so catch the error and ignore
                    pass

    def write_gromacs_file(self, input_type='input'):
        """To a gromacs file, write and format the necessary variables gro."""

        with open(f'{self.name}.gro', 'w+') as gro_file:
            gro_file.write(f'NEW {self.name.upper()} GRO FILE\n')
            gro_file.write(f'{len(self.coords[input_type]):>5}\n')
            for pos, atom in enumerate(self.coords[input_type], 1):
                # 'mol number''mol name'  'atom name'   'atom count'   'x coord'   'y coord'   'z coord'
                # 1WATER  OW1    1   0.126   1.624   1.679
                gro_file.write(
                    f'    1{self.name.upper()}  {atom[0]}{pos}   {pos}   '
                    f'{atom[1]: .3f}   {atom[2]: .3f}   {atom[3]: .3f}\n')

    def pickle(self, state=None):
        """
        Pickles the Molecule object in its current state to the (hidden) pickle file.
        If other pickle objects already exist for the particular object:
            the latest object is put to the top.
        """

        mols = OrderedDict()
        # First check if the pickle file exists
        try:
            # Try to load a hidden pickle file; make sure to get all objects
            with open('.QUBEKit_states', 'rb') as pickle_jar:
                while True:
                    try:
                        mol = pickle.load(pickle_jar)
                        mols[mol.state] = mol
                    except EOFError:
                        break
        except FileNotFoundError:
            pass

        # Now we can save the items; first assign the location
        self.state = state
        mols[self.state] = self

        # Open the pickle jar which will always be the ligand object's name
        with open(f'.QUBEKit_states', 'wb') as pickle_jar:

            # If there were other molecules of the same state in the jar: overwrite them
            for val in mols.values():
                pickle.dump(val, pickle_jar)

    def get_bond_equiv_classes(self):
        """
        Using the symmetry dict, give each bond a code. If any codes match, the bonds can be symmetrised.
        e.g. bond_symmetry_classes = {(0, 3): '2-0', (0, 4): '2-0', (0, 5): '2-0' ...}
        all of the above bonds (tuples) are of the same type (methyl H-C bonds in same region)
        This dict is then used to produce bond_types.
        bond_types is just a dict where the keys are the string code from above and the values are all
        of the bonds with that particular type.
        """

        bond_symmetry_classes = {}
        for bond in self.topology.edges:
            bond_symmetry_classes[bond] = (f'{self.atom_symmetry_classes[bond[0]]}-'
                                           f'{self.atom_symmetry_classes[bond[1]]}')

        self.bond_types = {}
        for key, val in bond_symmetry_classes.items():
            self.bond_types.setdefault(val, []).append(key)

    def get_angle_equiv_classes(self):
        """
        Using the symmetry dict, give each angle a code. If any codes match, the angles can be symmetrised.
        e.g. angle_symmetry_classes = {(1, 0, 3): '3-2-0', (1, 0, 4): '3-2-0', (1, 0, 5): '3-2-0' ...}
        all of the above angles (tuples) are of the same type (methyl H-C-H angles in same region)
        angle_types is just a dict where the keys are the string code from the above and the values are all
        of the angles with that particular type.
        """

        angle_symmetry_classes = {}
        for angle in self.angles:
            angle_symmetry_classes[angle] = (f'{self.atom_symmetry_classes[angle[0]]}-'
                                             f'{self.atom_symmetry_classes[angle[1]]}-'
                                             f'{self.atom_symmetry_classes[angle[2]]}')

        self.angle_types = {}
        for key, val in angle_symmetry_classes.items():
            self.angle_types.setdefault(val, []).append(key)

    def symmetrise_from_topo(self):
        """
        First, if rdkit_mol has been generated, get the bond and angle symmetry dicts.
        These will be used by L-J and the Harmonic Bond/Angle params

        Then, based on the molecule topology, symmetrise the methyl / amine hydrogens.
        If there's a carbon, does it have 3/2 hydrogens? -> symmetrise
        If there's a nitrogen, does it have 2 hydrogens? -> symmetrise
        Also keep a list of the methyl carbons and amine / nitrile nitrogens
        then exclude these bonds from the rotatable torsions list.
        """
        if self.rdkit_mol is not None:

            self.atom_symmetry_classes = RDKit().find_symmetry_classes(self.rdkit_mol)

            self.get_bond_equiv_classes()
            self.get_angle_equiv_classes()

        methyl_hs, amine_hs, other_hs = [], [], []
        methyl_amine_nitride_cores = []

        for atom in self.atoms:
            if atom.atomic_symbol == 'C' or atom.atomic_symbol == 'N':

                hs = []
                for bonded in self.topology.neighbors(atom.atom_index):
                    if len(list(self.topology.neighbors(bonded))) == 1:
                        # now make sure it is a hydrogen (as halogens could be caught here)
                        if self.atoms[bonded].atomic_symbol == 'H':
                            hs.append(bonded)

                if atom.atomic_symbol == 'C' and len(hs) == 2:    # This is part of a carbon hydrogen chain
                    other_hs.append(hs)
                elif atom.atomic_symbol == 'C' and len(hs) == 3:
                    methyl_hs.append(hs)
                    methyl_amine_nitride_cores.append(atom.atom_index)
                elif atom.atomic_symbol == 'N' and len(hs) == 2:
                    amine_hs.append(hs)
                    methyl_amine_nitride_cores.append(atom.atom_index)
                elif atom.atomic_symbol == 'N' and len(hs) == 1:
                    methyl_amine_nitride_cores.append(atom.atom_index)

        self.symm_hs = {'methyl': methyl_hs, 'amine': amine_hs, 'other': other_hs}

        # now modify the rotatable list to remove methyl and amine / nitrile torsions
        # these are already well represented in most FF's
        remove_list = []
        if self.rotatable is not None:
            rotatable = self.rotatable
            for key in rotatable:
                if key[0] in methyl_amine_nitride_cores or key[1] in methyl_amine_nitride_cores:
                    remove_list.append(key)

            # now remove the keys
            for torsion in remove_list:
                rotatable.remove(torsion)

            self.rotatable = rotatable if rotatable else None

    def openmm_coordinates(self, input_type='input'):
        """
        Take a set of coordinates from the molecule and convert them to openMM format
        :param input_type: The set of coordinates that should be used
        :return: A list of tuples of the coords
        """

        coordinates = self.coords[input_type]

        if input_type == 'traj' and len(coordinates) != len(self.coords['input']):
            # Multiple frames in this case
            return [[tuple(atom / 10) for atom in frame] for frame in coordinates]
        else:
            return [tuple(atom / 10) for atom in coordinates]

    def read_tdrive(self, bond_scan):
        """
        Read a tdrive qdata file and get the coordinates and scan energies and store in the molecule.
        :type bond_scan: the tuple of the scanned central bond
        :return: None, store the coords in the traj holder and the energies in the qm scan holder
        """

        scan_coords = []
        energy = []
        qm_scans = {}
        with open('qdata.txt', 'r') as data:
            for line in data.readlines():
                if 'COORDS' in line:
                    coords = [float(x) for x in line.split()[1:]]
                    coords = np.array(coords).reshape((len(self.atoms), 3))
                    scan_coords.append(coords)
                elif 'ENERGY' in line:
                    energy.append(float(line.split()[1]))

        qm_scans[bond_scan] = [np.array(energy), scan_coords]
        if qm_scans:
            self.qm_scans = qm_scans

    def read_scan_order(self, file):
        """
        Read a qubekit or tdrive dihedrals file and store the scan order into the ligand class
        :param file: The dihedrals input file.
        :return: The molecule with the scan_order saved
        """

        # If we have a QUBE_torsions.txt file get the scan order from there
        scan_order = []
        torsions = open(file).readlines()
        for line in torsions[2:]:
            torsion = line.split()
            if len(torsion) == 4:
                core = (int(torsion[1]), int(torsion[2]))
                if core in self.rotatable:
                    scan_order.append(core)
                elif reversed(tuple(core)) in self.rotatable:
                    scan_order.append(reversed(tuple(core)))

        self.scan_order = scan_order


class Ligand(DefaultsMixin, Molecule):

    def __init__(self, mol_input, name=None):
        """
        parameter_engine        A string keeping track of the parameter engine used to assign the initial parameters
        hessian                 2d numpy array; matrix of size 3N x 3N where N is number of atoms in the molecule
        modes                   A list of the qm predicted frequency modes
        home

        constraints_file        Either an empty string (does nothing in geometric run command); or
                                the abspath of the constraint.txt file (constrains the execution of geometric)
        """

        super().__init__(mol_input, name)

        self.parameter_engine = 'openmm'
        self.hessian = None
        self.modes = None
        self.home = None

        self.constraints_file = None

        # Make sure we have the topology before we calculate the properties
        if self.topology.edges:
            self.find_angles()
            self.find_dihedrals()
            self.find_rotatable_dihedrals()
            self.find_impropers()
            self.get_dihedral_values()
            self.find_bond_lengths()
            self.get_angle_values()
            self.symmetrise_from_topo()

    def write_pdb(self, input_type='input', name=None):
        """
        Take the current molecule and topology and write a pdb file for the molecule.
        Only for small molecules, not standard residues. No size limit.
        """

        molecule = self.coords[input_type]

        with open(f'{name if name is not None else self.name}.pdb', 'w+') as pdb_file:

            # Write out the atomic xyz coordinates
            pdb_file.write(f'REMARK   1 CREATED WITH QUBEKit {datetime.now()}\n')
            pdb_file.write(f'COMPND    {self.name:<20}\n')
            for i, atom in enumerate(molecule):
                pdb_file.write(
                    f'HETATM {i+1:>4}{self.atoms[i].atom_name:>4}  UNL     1{atom[0]:12.3f}{atom[1]:8.3f}{atom[2]:8.3f}'
                    f'  1.00  0.00         {self.atoms[i].atomic_symbol.upper():>3}\n')

            # Now add the connection terms
            for node in self.topology.nodes:
                bonded = sorted(list(nx.neighbors(self.topology, node)))
                if len(bonded) > 1:
                    pdb_file.write(f'CONECT{node + 1:5}{"".join(f"{x + 1:5}" for x in bonded)}\n')

            pdb_file.write('END\n')


class Protein(DefaultsMixin, Molecule):
    """This class handles the protein input to make the qubekit xml files and rewrite the pdb so we can use it."""

    def __init__(self, filename):

        super().__init__(filename)

        self.pdb_names = None
        # TODO Needs updating with new Path method of handling filenames
        self.read_pdb(self.filename)
        self.residues = None
        self.home = os.getcwd()

    def read_pdb(self, input_type='input'):
        """
        Read the pdb file which probably does not have the right connections,
        so we need to find them using QUBE.xml
        """

        with open(self.filename, 'r') as pdb:
            lines = pdb.readlines()

        protein = []
        self.topology = nx.Graph()
        self.residues = []
        self.Residues = []
        self.pdb_names = []
        self.atoms = []

        # atom counter used for graph node generation
        atom_count = 0
        for line in lines:
            if 'ATOM' in line or 'HETATM' in line:
                atomic_symbol = str(line[76:78])
                atomic_symbol = re.sub('[0-9]+', '', atomic_symbol).strip()

                # If the element column is missing from the pdb, extract the atomic_symbol from the atom name.
                if not atomic_symbol:
                    atomic_symbol = str(line.split()[2])
                    atomic_symbol = re.sub('[0-9]+', '', atomic_symbol)

                # now make sure we have a valid element
                if atomic_symbol.lower() == 'cl' or atomic_symbol.lower() == 'br':
                    pass
                else:
                    atomic_symbol = atomic_symbol[0]

                atom_name = f'{atomic_symbol}{atom_count}'
                qube_atom = Atom(Element().number(atomic_symbol), atom_count, atom_name)

                self.atoms.append(qube_atom)

                self.pdb_names.append(str(line.split()[2]))

                # also get the residue order from the pdb file so we can rewrite the file
                self.Residues.append(str(line.split()[3]))

                # Also add the atom number as the node in the graph
                self.topology.add_node(atom_count)
                atom_count += 1
                protein.append([float(line[30:38]), float(line[38:46]), float(line[46:54])])

            elif 'CONECT' in line:
                # Now look through the connectivity section and add all edges to the graph corresponding to the bonds.
                for i in range(2, len(line.split())):
                    if int(line.split()[i]) != 0:
                        self.topology.add_edge(int(line.split()[1]) - 1, int(line.split()[i]) -1)

        self.coords[input_type] = np.array(protein)

        # check if there are any conect terms in the file first
        # if not self.topology.edges:
        if not len(self.topology.edges):
            print('No connections found!')
        else:
            self.find_angles()
            self.find_dihedrals()
            self.find_rotatable_dihedrals()
            self.find_impropers()
            self.get_dihedral_values()
            self.find_bond_lengths()
            self.get_angle_values()
            self.symmetrise_from_topo()

        # TODO What if there are two or more of the same residue back to back?
        #   Need to store the number of atoms in each amino acid and use that to check instead.
        # Remove duplicates
        self.residues = [res for res, group in groupby(self.Residues)]

    def write_pdb(self, name=None):
        """This method replaces the ligand method as all of the atom names and residue names have to be replaced."""

        molecule = self.coords['input']

        with open(f'{name if name is not None else self.name}.pdb', 'w+') as pdb_file:

            # Write out the atomic xyz coordinates
            pdb_file.write(f'REMARK   1 CREATED WITH QUBEKit {datetime.now()}\n')
            # pdb_file.write(f'COMPND    {self.name:<20}\n')
            # we have to transform the atom name while writing out the pdb file
            for i, atom in enumerate(molecule):
                pdb_file.write(
                    f'HETATM {i+1:>4}{self.atoms[i].atom_name:>4}  QUP     1{atom[0]:12.3f}{atom[1]:8.3f}{atom[2]:8.3f}'
                    f'  1.00  0.00         {self.atoms[i].atomic_symbol.upper():>3}\n')

            # Now add the connection terms
            for node in self.topology.nodes:
                bonded = sorted(list(nx.neighbors(self.topology, node)))
                if len(bonded) >= 1:
                    pdb_file.write(f'CONECT{node + 1:5}{"".join(f"{x + 1:5}" for x in bonded)}\n')

            pdb_file.write('END\n')

    def update(self, input_type='input'):
        """
        After the protein has been passed to the parametrisation class we get back the bond info
        use this to update all missing terms.
        """

        # using the new harmonic bond force dict we can add the bond edges to the topology graph
        for bond in self.HarmonicBondForce:
            self.topology.add_edge(*bond)

        self.find_angles()
        self.find_dihedrals()
        self.find_rotatable_dihedrals()
        self.get_dihedral_values(input_type)
        self.find_bond_lengths(input_type)
        self.get_angle_values(input_type)
        self.find_impropers()
        # this creates the dictionary of terms that should be symmetrise
        self.symmetrise_from_topo()



#!/usr/bin/env python3

# TODO
#  Squash unnecessary arguments into self.molecule. Args such as torsion_options.
#  Better handling of torsion_options
#  Option to use numbers to skip e.g. -skip 4 5 : skips hessian and mod_seminario steps

from QUBEKit.dihedrals import TorsionScan, TorsionOptimiser
from QUBEKit.engines import PSI4, Chargemol, Gaussian, ONETEP, QCEngine, RDKit
from QUBEKit.lennard_jones import LennardJones
from QUBEKit.ligand import Ligand
from QUBEKit.mod_seminario import ModSeminario
from QUBEKit.parametrisation import OpenFF, AnteChamber, XML
from QUBEKit.parametrisation.base_parametrisation import Parametrisation
from QUBEKit.utils import constants
from QUBEKit.utils.decorators import exception_logger
from QUBEKit.utils.exceptions import OptimisationFailed, HessianCalculationFailed
from QUBEKit.utils.helpers import mol_data_from_csv, generate_bulk_csv, append_to_log, pretty_progress, pretty_print, \
    Configure, unpickle

import argparse
from collections import OrderedDict
from datetime import datetime
from functools import partial
import os
from shutil import copy, move
import subprocess as sp
import sys

import numpy as np


# To avoid calling flush=True in every print statement.
printf = partial(print, flush=True)


class ArgsAndConfigs:
    """
    This class will be called once for any individual or bulk run.
    As such, this class needs to:
        * Parse all commands from terminal
        * If individual:
            * Initialise molecule
            * Using proper config file (from terminal command or default)
            * Store all configs from file into Molecule object
            * Store any config changes from terminal into Molecule object
            * Call Execute (which handles: working dir, logging, order, etc; see docstring for info)

        * If bulk:
            * For each molecule:
                * Initialise molecule
                * Using proper config file (from terminal command or default)
                * Store all configs from file
                * Store any config changes from terminal
                * Call Execute
    """

    def __init__(self):
        # First make sure the config folder has been made missing for conda and pip
        home = os.path.expanduser('~')
        config_folder = os.path.join(home, 'QUBEKit_configs')
        if not os.path.exists(config_folder):
            os.makedirs(config_folder)
            printf(f'Making config folder at: {home}')

        self.args = self.parse_commands()

        # If we are doing the torsion test add the attribute to the molecule so we can catch it in execute
        if self.args.torsion_test:
            self.args.restart = 'finalise'

        # If it's a bulk run, handle it separately
        # TODO Add .sdf as possible bulk_run, not just .csv
        if self.args.bulk_run is not None:
            self.handle_bulk()

        elif self.args.restart is not None:
            # Find the pickled checkpoint file and load it as the molecule
            try:
                self.molecule = unpickle()[self.args.restart]
            except KeyError:
                raise KeyError('This stage was not found in the log file; was the previous stage completed?')
        else:
            # Initialise molecule
            if self.args.smiles:
                self.molecule = Ligand(*self.args.smiles)
                # Now we should create the initial molecule and
            else:
                self.molecule = Ligand(self.args.input)

        # Find which config file is being used
        self.molecule.config_file = self.args.config_file

        # Handle configs which are in a file
        file_configs = Configure().load_config(self.molecule.config_file)
        for name, val in file_configs.items():
            setattr(self.molecule, name, val)

        # Although these may be None always, they need to be explicitly set anyway.
        self.molecule.restart = None
        self.molecule.end = None
        self.molecule.skip = None

        # Handle configs which are changed by terminal commands
        for name, val in vars(self.args).items():
            if val is not None:
                setattr(self.molecule, name, val)

        # Now we need to remove torsion_test as it is passed from the command line
        if self.args.torsion_test is False:
            del self.molecule.torsion_test

        # Now check if we have been supplied a dihedral file and a constraints file
        if self.args.dihedral_file:
            self.molecule.read_scan_order(self.args.dihedral_file)
        if self.args.constraints_file:
            self.molecule.constraints_file = self.args.constaints_file

        # If restarting put the molecule back into the checkpoint file with the new configs
        if self.args.restart is not None:
            self.molecule.pickle(state=self.args.restart)

        # Now that all configs are stored correctly: execute.
        Execute(self.molecule)

    def __repr__(self):
        return f'{self.__class__.__name__}({self.__dict__!r})'

    @staticmethod
    def parse_commands():
        """
        Parses commands from the terminal using argparse.

        Contains classes for handling actions as well as simple arg parsers for config changes.
        """

        # Action classes
        class SetupAction(argparse.Action):
            """The setup action class that is called when setup is found in the command line."""

            def __call__(self, pars, namespace, values, option_string=None):
                """This function is executed when setup is called."""

                choice = int(input('You can now edit config files using QUBEKit, choose an option to continue:\n'
                                   '1) Edit a config file\n'
                                   '2) Create a new master template\n'
                                   '3) Make a normal config file\n'
                                   '4) Cancel\n>'))

                if choice == 1:
                    inis = Configure().show_ini()
                    name = input(f'Enter the name or number of the config file to edit\n'
                                 f'{"".join(f"{inis.index(ini)}:{ini}    " for ini in inis)}\n>')
                    # make sure name is right
                    if name in inis:
                        Configure().ini_edit(name)
                    else:
                        Configure().ini_edit(inis[int(name)])

                elif choice == 2:
                    Configure().ini_writer('master_config.ini')
                    Configure().ini_edit('master_config.ini')

                elif choice == 3:
                    name = input('Enter the name of the config file to create\n>')
                    Configure().ini_writer(name)
                    Configure().ini_edit(name)

                else:
                    sys.exit('Cancelling setup; no changes made. '
                             'If you accidentally entered the wrong key, restart with QUBEKit -setup')

                sys.exit()

        class CSVAction(argparse.Action):
            """The csv creation class run when the csv option is used."""

            def __call__(self, pars, namespace, values, option_string=None):
                """This function is executed when csv is called."""

                generate_bulk_csv(*values)
                sys.exit()

        class ProgressAction(argparse.Action):
            """Run the pretty progress function to get the progress of all running jobs."""

            def __call__(self, pars, namespace, values, option_string=None):
                """This function is executed when progress is called."""

                pretty_progress()
                sys.exit()

        class TorsionMakerAction(argparse.Action):
            """Help the user make a torsion scan file."""

            def __call__(self, pars, namespace, values, option_string=None):
                """This function is executed when Torsion maker is called."""

                # load in the ligand molecule
                mol = Ligand(values)

                # Prompt the user for the scan order
                scanner = TorsionScan(mol)
                scanner.find_scan_order()

                # Write out the scan file
                with open(f'{mol.name}.dihedrals', 'w+') as qube:
                    qube.write('# dihedral definition by atom indices starting from 0\n#  i      j      k      l\n')
                    for scan in mol.scan_order:
                        scan_di = mol.dihedrals[scan][0]
                        qube.write(f'  {scan_di[0]:2}     {scan_di[1]:2}     {scan_di[2]:2}     {scan_di[3]:2}\n')
                printf(f'{mol.name}.dihedrals made.')

                sys.exit()

        def string_to_bool(string):
            """Convert a string to a bool for argparse use when casting to bool"""
            return string.lower() in ['true', 't', 'yes', 'y']

        intro = 'Welcome to QUBEKit! For a list of possible commands, use the help command: -h.' \
                'Alternatively, take a look through our github page for commands, recipes and common problems:' \
                'http://github.com/qubekit/QUBEKit'
        parser = argparse.ArgumentParser(prog='QUBEKit', formatter_class=argparse.RawDescriptionHelpFormatter,
                                         description=intro)

        # Add all of the command line options in the arg parser
        parser.add_argument('-c', '--charge', type=int, help='Enter the charge of the molecule, default 0.')
        parser.add_argument('-m', '--multiplicity', type=int,
                            help='Enter the multiplicity of the molecule, default 1.')
        parser.add_argument('-threads', '--threads', type=int,
                            help='Number of threads used in various stages of analysis, especially for engines like '
                                 'PSI4, Gaussian09, etc. Value is given as an int.')
        parser.add_argument('-memory', '--memory', type=int,
                            help='Amount of memory used in various stages of analysis, especially for engines like '
                                 'PSI4, Gaussian09, etc. Value is given as an int, e.g. 6GB is simply 6.')
        parser.add_argument('-ddec', '--ddec_version', choices=[3, 6], type=int,
                            help='Enter the ddec version for charge partitioning, does not effect ONETEP partitioning.')
        parser.add_argument('-geo', '--geometric', choices=[True, False], type=string_to_bool,
                            help='Turn on geometric to use this during the qm optimisations, recommended.')
        parser.add_argument('-bonds', '--bonds_engine', choices=['psi4', 'g09', 'g16'],
                            help='Choose the QM code to calculate the bonded terms.')
        parser.add_argument('-charges', '--charges_engine', choices=['onetep', 'chargemol'],
                            help='Choose the method to do the charge partitioning.')
        parser.add_argument('-density', '--density_engine', choices=['onetep', 'g09', 'g16', 'psi4'],
                            help='Enter the name of the QM code to calculate the electron density of the molecule.')
        parser.add_argument('-solvent', '--solvent', choices=[True, False], type=string_to_bool,
                            help='Enter whether or not you would like to use a solvent.')
        # Maybe separate into known solvents and IPCM constants?
        parser.add_argument('-convergence', '--convergence', choices=['GAU', 'GAU_TIGHT', 'GAU_VERYTIGHT'],
                            type=str.upper, help='Enter the convergence criteria for the optimisation.')
        parser.add_argument('-param', '--parameter_engine', choices=['xml', 'antechamber', 'openff', 'none'],
                            help='Enter the method of where we should get the initial molecule parameters from, '
                                 'if xml make sure the xml has the same name as the pdb file.')
        parser.add_argument('-mm', '--mm_opt_method', choices=['openmm', 'rdkit_mff', 'rdkit_uff', 'none'],
                            help='Enter the mm optimisation method for pre qm optimisation.')
        parser.add_argument('-config', '--config_file', default='default_config', choices=Configure().show_ini(),
                            help='Enter the name of the configuration file you wish to use for this run from the list '
                                 'available, defaults to master.')
        parser.add_argument('-theory', '--theory',
                            help='Enter the name of the qm theory you would like to use.')
        parser.add_argument('-basis', '--basis',
                            help='Enter the basis set you would like to use.')
        parser.add_argument('-restart', '--restart', choices=['parametrise', 'mm_optimise', 'qm_optimise', 'hessian',
                                                              'mod_sem', 'density', 'charges', 'lennard_jones',
                                                              'torsion_scan', 'torsion_optimise'],
                            help='Enter the restart point of a QUBEKit job.')
        parser.add_argument('-end', '-end', choices=['mm_optimise', 'qm_optimise', 'hessian', 'mod_sem', 'density',
                                                     'charges', 'lennard_jones', 'torsion_scan', 'torsion_optimise',
                                                     'finalise'],
                            help='Enter the end point of the QUBEKit job.')
        parser.add_argument('-progress', '--progress', nargs='?', const=True,
                            help='Get the current progress of a QUBEKit single or bulk job.', action=ProgressAction)
        parser.add_argument('-skip', '--skip', nargs='+', choices=['mm_optimise', 'qm_optimise', 'hessian', 'mod_sem',
                                                                   'density', 'charges', 'lennard_jones',
                                                                   'torsion_scan', 'torsion_optimise', 'finalise'],
                            help='Option to skip certain stages of the execution.')
        parser.add_argument('-tor_test', '--torsion_test', action='store_true',
                            help='Enter True if you would like to run a torsion test on the chosen torsions.')
        parser.add_argument('-tor_make', '--torsion_maker', action=TorsionMakerAction,
                            help='Allow QUBEKit to help you make a torsion input file for the given molecule')
        parser.add_argument('-log', '--log', type=str,
                            help='Enter a name to tag working directories with. Can be any alphanumeric string.'
                                 'This helps differentiate (by more than just date) different analyses of the '
                                 'same molecule.')
        parser.add_argument('-vib', '--vib_scaling', type=float,
                            help='Enter the vibrational scaling to be used with the basis set.')
        parser.add_argument('-iters', '--iterations', type=int,
                            help='Max number of iterations for QM scan.')
        parser.add_argument('-constraints', '--constraints_file', type=str,
                            help='The name of the geometric constraints file.')
        parser.add_argument('-dihedrals', '--dihedral_file', type=str,
                            help='The name of the qubekit/tdrive torsion file.')

        # Add mutually exclusive groups to stop certain combinations of options,
        # e.g. setup should not be run with csv command
        groups = parser.add_mutually_exclusive_group()
        groups.add_argument('-setup', '--setup_config', nargs='?', const=True,
                            help='Setup a new configuration or edit an existing one.', action=SetupAction)
        groups.add_argument('-sm', '--smiles', nargs='+',
                            help='Enter the smiles string of a molecule as a starting point.')
        groups.add_argument('-bulk', '--bulk_run',
                            help='Enter the name of the csv file to run as bulk, bulk will use smiles unless it finds '
                                 'a molecule file with the same name.')
        groups.add_argument('-csv', '--csv_filename', action=CSVAction, nargs='*',
                            help='Enter the name of the csv file you would like to create for bulk runs.'
                                 'Optionally, you may also add the maximum number of molecules per file.')
        groups.add_argument('-i', '--input', help='Enter the molecule input pdb file (only pdb so far!)')
        groups.add_argument('-version', '--version', action='version', version='2.5.1')

        return parser.parse_args()

    def handle_bulk(self):
        """
        Getting and setting configs for bulk runs is a little different, requiring this method.
        The configs are taken from the .csv, then the .ini, then the terminal.
        This is repeated for each molecule in the bulk run, then Execute is called.

        Configs cannot be changed between molecule analyses as config data is
        only loaded once at the start; -restart is required for that.
        """

        csv_file = self.args.bulk_run
        # mol_data_from_csv handles defaults if no argument is given
        bulk_data = mol_data_from_csv(csv_file)

        names = list(bulk_data)

        home = os.getcwd()

        for name in names:
            printf(f'Analysing: {name}\n')

            # Get pdb from smiles or name if no smiles is given
            if bulk_data[name]['smiles'] is not None:
                smiles_string = bulk_data[name]['smiles']
                self.molecule = Ligand(smiles_string, name)

            else:
                # TODO Different file types
                # Initialise molecule, ready to add configs to it
                self.molecule = Ligand(f'{name}.pdb')

            # Read each row in bulk data and set it to the molecule object
            for key, val in bulk_data[name].items():
                setattr(self.molecule, key, val)

            self.molecule.skip = None

            # Using the config file from the .csv, gather the .ini file configs
            file_configs = Configure().load_config(self.molecule.config_file)
            for key, val in file_configs.items():
                setattr(self.molecule, key, val)

            # Handle configs which are changed by terminal commands
            for key, val in vars(self.args).items():
                if val is not None:
                    setattr(self.molecule, key, val)

            # Now that all configs are stored correctly: execute.
            Execute(self.molecule)

            os.chdir(home)

        sys.exit('Bulk analysis complete.\nUse QUBEKit -progress to view the completion progress of your molecules')


class Execute:
    """
    This class will be called for each INDIVIDUAL run and for each analysis in a BULK run.
    ArgsAndConfigs will handle bulk runs, calling this class as and when necessary.

    As such, this class needs to handle for each run:
        * Create and / or move into working dir
        * Create log file (Skipped if restarting)
        * Write to log file what is happening at the start of execution
            (starting or continuing; what are the configs; etc)
        * Execute any stages required, in the correct order according to self.order
        * Store info to log file as the program runs
            (which stages are being called; any key info / timings regarding progress; errors; etc)
        * Return results, both in individual stage directories and at the end
    """

    def __init__(self, molecule):

        # At this point, molecule should contain all of the config options from
        # the defaults, config file and terminal commands. These all come from the ArgsAndConfigs class.
        self.molecule = molecule

        self.start_up_msg = ('If QUBEKit ever breaks or you would like to view timings and loads of other info, '
                             'view the log file.\nOur documentation (README.md) '
                             'also contains help on handling the various commands for QUBEKit.\n')

        self.order = OrderedDict([('parametrise', self.parametrise),
                                  ('mm_optimise', self.mm_optimise),
                                  ('qm_optimise', self.qm_optimise),
                                  ('hessian', self.hessian),
                                  ('mod_sem', self.mod_sem),
                                  ('density', self.density),
                                  ('charges', self.charges),
                                  ('lennard_jones', self.lennard_jones),
                                  ('torsion_scan', self.torsion_scan),
                                  ('torsion_optimise', self.torsion_optimise),
                                  ('finalise', self.finalise),
                                  ('torsion_test', self.torsion_test)])

        # Keep this for pickling
        self.immutable_order = list(self.order)

        self.engine_dict = {'psi4': PSI4, 'g09': Gaussian, 'g16': Gaussian, 'onetep': ONETEP}

        printf(self.start_up_msg)

        # If restart is None, then the analysis has not been started previously
        self.create_log() if self.molecule.restart is None else self.continue_log()

        self.redefine_order()

        self.run()

    def __repr__(self):
        return f'{self.__class__.__name__}({self.__dict__!r})'

    def redefine_order(self):
        """
        If any order changes are required (restarting, new end point, skipping stages), it is done here.
        Creates a new self.order based on self.molecule's configs.
        """

        # If we are doing a torsion_test we have to redo the order as follows:
        # 1 Skip the finalise step to create a pickled ligand at the torsion_test stage
        # 2 Do the torsion_test and delete the torsion_test attribute
        # 3 Do finalise again to save the ligand with the correct attributes
        if hasattr(self.molecule, 'torsion_test'):
            self.order = OrderedDict([('finalise', self.skip), ('torsion_test', self.torsion_test), ('finalise', self.skip)])

        else:
            start = self.molecule.restart if self.molecule.restart is not None else 'parametrise'
            end = self.molecule.end if self.molecule.end is not None else 'finalise'
            skip = self.molecule.skip if self.molecule.skip is not None else []

            # Create list of all keys
            stages = list(self.order)

            # Cut out the keys before the start_point and after the end_point
            # Add finalise back in if it's removed (finalise should always be called).
            stages = stages[stages.index(start):stages.index(end) + 1] + ['finalise']

            # Redefine self.order to only contain the key, val pairs from stages
            self.order = OrderedDict(pair for pair in self.order.items() if pair[0] in set(stages))

            for pair in self.order.items():
                if pair[0] in skip:
                    self.order[pair[0]] = self.skip
                else:
                    self.order[pair[0]] = pair[1]

    def create_log(self):
        """
        Creates the working directory for the job as well as the log file.
        This log file is then extended when:
            - decorators.timer_logger wraps a called method;
            - helpers.append_to_log() is called;
            - helpers.pretty_print() is called with to_file set to True;
            - decorators.exception_logger() wraps a function / method which throws an exception.
                (This captures almost all errors)

        This method also makes backups if the working directory already exists.
        """

        date = datetime.now().strftime('%Y_%m_%d')

        # Define name of working directory.
        # This is formatted as 'QUBEKit_molecule name_yyyy_mm_dd_log_string'.

        dir_name = f'QUBEKit_{self.molecule.name}_{date}_{self.molecule.log}'

        # If you can't make a dir because it exists, back it up.
        try:
            os.mkdir(dir_name)

        except FileExistsError:
            # Try make a backup folder
            try:
                os.mkdir('QUBEKit_backups')
                printf('Making backup folder: QUBEKit_backups')
            except FileExistsError:
                # Backup folder already made
                pass
            # Keep increasing backup number until that particular number does not exist
            finally:
                count = 1
                while os.path.exists(f'QUBEKit_backups/{dir_name}_{str(count).zfill(3)}'):
                    count += 1

                # Then, make that backup and make a new working directory
                move(dir_name, f'QUBEKit_backups/{dir_name}_{str(count).zfill(3)}')
                printf(f'Moving directory: {dir_name} to backup folder')
                os.mkdir(dir_name)

        # Finally, having made any necessary backups, move files and change to working dir.
        finally:
            os.chdir(dir_name)

            # Set a home directory
            self.molecule.home = os.getcwd()

        # Find external files
        copy_files = [f'{self.molecule.name}.xml', self.molecule.filename]
        for file in copy_files:
            try:
                copy(f'../{file}', file)
            except FileNotFoundError:
                pass
            except TypeError:
                pass

        with open('QUBEKit_log.txt', 'w+') as log_file:
            log_file.write(f'Beginning log file; the time is: {datetime.now()}\n\n\n')

        self.log_configs()

    def continue_log(self):
        """
        In the event of restarting an analysis, find and append to the existing log file
        rather than creating a new one.
        """

        with open('QUBEKit_log.txt', 'a+') as log_file:
            log_file.write(f'\n\nContinuing log file from previous execution; the time is: {datetime.now()}\n\n\n')

        self.log_configs()

    def log_configs(self):
        """Writes the runtime and file-based defaults to a log file."""

        with open('QUBEKit_log.txt', 'a+') as log_file:

            log_file.write(f'Analysing: {self.molecule.name}\n\n')

            log_file.write('The runtime defaults and config options are as follows:\n\n')
            for key, val in self.molecule.__dict__.items():
                if val is not None:
                    log_file.write(f'{key}: {val}\n')
            log_file.write('\n')

    @exception_logger
    def run(self, torsion_options=None):
        # TODO bulk torsion options need to be checked
        """
        Calls all the relevant classes and methods for the full QM calculation in the correct order
            (according to self.order).
        Exceptions are added to log (if raised) using the decorator.
        """

        if 'parametrise' in self.order:
            if torsion_options is not None:
                torsion_options = torsion_options.split(',')
                self.molecule = self.store_torsions(self.molecule, torsion_options)
            self.molecule.pickle(state='parametrise')

        stage_dict = {
            'parametrise': ['Parametrising molecule', 'Molecule parametrised'],
            'mm_optimise': ['Partially optimising with MM', 'Partial optimisation complete'],
            'qm_optimise': ['Optimising molecule, view .xyz file for progress', 'Molecule optimisation complete'],
            'hessian': ['Calculating Hessian matrix', 'Hessian matrix calculated and confirmed to be symmetric'],
            'mod_sem': ['Calculating bonds and angles with modified Seminario method', 'Bonds and angles calculated'],
            'density': [f'Performing density calculation with {self.molecule.density_engine}',
                        'Density calculation complete'],
            'charges': [f'Chargemol calculating charges using DDEC{self.molecule.ddec_version}', 'Charges calculated'],
            'lennard_jones': ['Performing Lennard-Jones calculation', 'Lennard-Jones parameters calculated'],
            'torsion_scan': ['Performing QM-constrained optimisation with Torsiondrive',
                             'Torsiondrive finished and QM results saved'],
            'torsion_optimise': ['Performing torsion optimisation', 'Torsion optimisation complete'],
            'finalise': ['Finalising analysis', 'Molecule analysis complete!'],
            'pause': ['Pausing analysis', 'Analysis paused!'],
            'skip': ['Skipping section', 'Section skipped'],
            'torsion_test': ['Testing torsion single point energies', 'Torsion testing complete']}

        # Do the first stage in the order to get the next_key for the following loop
        key = list(self.order)[0]
        next_key = self.stage_wrapper(key, *stage_dict[key], torsion_options)

        # Cannot use for loop as we mutate the dictionary during the loop
        while True:
            if next_key is None:
                break
            else:
                next_key = self.stage_wrapper(next_key, *stage_dict[next_key])

            if next_key == 'pause':
                self.pause()
                break

    def stage_wrapper(self, start_key, begin_log_msg='', fin_log_msg='', torsion_options=None):
        """
        Firstly, check if the stage start_key is in self.order; this tells you if the stage should be called or not.
        If it isn't in self.order:
            - Do nothing
        If it is:
            - Unpickle the ligand object at the start_key stage
            - Write to the log that something's about to be done (if specified)
            - Make (if not restarting) and / or move into the working directory for that stage
            - Do the thing
            - Move back out of the working directory for that stage
            - Write to the log that something's been done (if specified)
            - Pickle the ligand object again with the next_key marker as its stage
        """

        mol = unpickle()[start_key]

        # Set the state for logging any exceptions should they arise
        mol.state = start_key

        # if we have a torsion options dictionary pass it to the molecule
        if torsion_options is not None:
            mol = self.store_torsions(mol, torsion_options)

        skipping = False
        if self.order[start_key] == self.skip:
            printf(f'Skipping stage: {start_key}')
            append_to_log(f'skipping stage: {start_key}')
            skipping = True
        else:
            if begin_log_msg:
                printf(f'{begin_log_msg}...', end=' ')

        home = os.getcwd()

        folder_name = f'{str(self.immutable_order.index(start_key) + 1).zfill(2)}_{start_key}'
        # Make sure you don't get an error if restarting
        try:
            os.mkdir(folder_name)
        except FileExistsError:
            pass
        finally:
            os.chdir(folder_name)

        self.order[start_key](mol)
        self.order.pop(start_key, None)
        os.chdir(home)

        # Begin looping through self.order, but return after the first iteration.
        for key in self.order:
            next_key = key
            if fin_log_msg and not skipping:
                printf(fin_log_msg)

            mol.pickle(state=next_key)
            return next_key

    @staticmethod
    def parametrise(molecule):
        """Perform initial molecule parametrisation using OpenFF, Antechamber or XML."""

        append_to_log('Starting parametrisation')

        # Parametrisation options:
        param_dict = {'antechamber': AnteChamber, 'xml': XML, 'openff': OpenFF, 'none': Parametrisation}

        # If we are using xml we have to move it to QUBEKit working dir
        if molecule.parameter_engine == 'xml':
            try:
                copy(os.path.join(molecule.home, f'{molecule.name}.xml'), f'{molecule.name}.xml')
            except FileNotFoundError:
                raise FileNotFoundError('You need to supply an xml file if you wish to use xml-based parametrisation;\n'
                                        'put this file in the location you are running QUBEKit from.\n'
                                        'Alternatively, use a different parametrisation method such as:\n'
                                        '-param antechamber')

        # Perform the parametrisation
        # If the method is none the molecule is not parameterised but the parameter holders are initiated
        if molecule.parameter_engine == 'none':
            param_dict[molecule.parameter_engine](molecule).gather_parameters()
        else:
            # Write the PDB file; this covers us if we have a different input file
            molecule.write_pdb()
            param_dict[molecule.parameter_engine](molecule)

        append_to_log(f'Finishing parametrisation of molecule with {molecule.parameter_engine}')

        return molecule

    @staticmethod
    def mm_optimise(molecule):
        """
        Use an mm force field to get the initial optimisation of a molecule

        options
        ---------
        RDKit MFF or UFF force fields can have strange effects on the geometry of molecules

        Geometric / OpenMM depends on the force field the molecule was parameterised with gaff/2, OPLS smirnoff.
        """

        append_to_log('Starting mm_optimisation')
        # Check which method we want then do the optimisation
        if molecule.mm_opt_method == 'none':
            # Skip the optimisation step
            molecule.coords['mm'] = molecule.coords['input']

        elif molecule.mm_opt_method == 'openmm':
            if molecule.parameter_engine != 'none':
                # Make the inputs
                molecule.write_pdb(input_type='input')
                molecule.write_parameters()
                # Run geometric
                # TODO Should this be moved to allow a decorator?
                with open('log.txt', 'w+') as log:
                    sp.run(f'geometric-optimize --reset --epsilon 0.0 --maxiter {molecule.iterations} --pdb '
                           f'{molecule.name}.pdb --openmm {molecule.name}.xml '
                           f'{molecule.constraints_file if molecule.constraints_file is not None else ""}',
                           shell=True, stdout=log, stderr=log)

                # This will continue even if we don't converge this is fine
                # Read the xyz traj and store the frames
                molecule.read_file(f'{molecule.name}_optim.xyz', input_type='traj')
                # Store the last from the traj as the mm optimised structure
                molecule.coords['mm'] = molecule.coords['traj'][-1]
            else:
                raise OptimisationFailed('You can not optimise a molecule with OpenMM and no initial parameters; '
                                         'consider parametrising or using UFF/MFF in RDKit')

        else:
            # TODO change to qcengine as this can already be done
            # Run an rdkit optimisation with the right FF
            rdkit_ff = {'rdkit_mff': 'MFF', 'rdkit_uff': 'UFF'}
            molecule.filename = RDKit().mm_optimise(molecule.filename, ff=rdkit_ff[molecule.mm_opt_method])

        append_to_log(f'Finishing mm_optimisation of the molecule with {molecule.mm_opt_method}')

        return molecule

    def qm_optimise(self, molecule):
        """Optimise the molecule with or without geometric."""

        # TODO This method's not always printing completion to log file.
        # TODO If the optimisation restarts, we should print that to the log file.
        #   Also, restarts should restart from the partially optimised structure (last trajectory before failure)

        append_to_log('Starting qm_optimisation')
        qm_engine = self.engine_dict[molecule.bonds_engine](molecule)

        if molecule.geometric and molecule.bonds_engine == 'psi4':

            qceng = QCEngine(molecule)
            # See if the structure is there if not we did not optimise
            result = qceng.call_qcengine('geometric', 'gradient',
                                         input_type=f'{"mm" if list(molecule.coords["mm"]) else "input"}')

            # Check if converged and get the geometry
            if result['success']:

                # Load all of the frames into the molecule's trajectory holder
                molecule.read_geometric_traj(result['trajectory'])

                # store the final molecule as the qm optimised structure
                molecule.coords['qm'] = np.array(result['final_molecule']['geometry']).reshape((len(molecule.atoms), 3))
                molecule.coords['qm'] *= constants.BOHR_TO_ANGS

                molecule.qm_energy = result['energies'][-1]

                # Write out the trajectory file
                molecule.write_xyz(input_type='traj', name=f'{molecule.name}_opt')
                molecule.write_xyz(input_type='qm', name='opt')

                append_to_log(f'Finishing qm_optimisation of molecule{" using geometric" if molecule.geometric else ""}')

                return molecule

            else:
                restart_count = 1
                while not result['success'] and restart_count < 3:

                    # If the error was not a straight segfault se if we can get the molecule so far
                    try:
                        molecule.coords['temp'] = np.array(result['input_data']['final_molecule']['geometry']).reshape((len(molecule.atoms), 3))
                        molecule.coords['temp'] *= constants.BOHR_TO_ANGS

                        result = qceng.call_qcengine('geometric', 'gradient', 'temp')

                    except KeyError:
                        # No molecule data was caught so start again
                        # TODO check for other errors here as well
                        result = qceng.call_qcengine(
                            'geometric', 'gradient', f'{"mm" if list(molecule.coords["mm"]) else "input"}')

                if result['success']:
                    molecule.coords['qm'] = np.array(result['final_molecule']['geometry']).reshape((len(molecule.atoms), 3))
                    molecule.coords['qm'] *= constants.BOHR_TO_ANGS

                    molecule.qm_energy = result['energies'][-1]

                    return molecule

                else:
                    raise OptimisationFailed("The optimisation did not converge")

        result = qm_engine.generate_input(input_type=f'{"mm" if list(molecule.coords["mm"]) else "input"}',
                                          optimise=True, execute=molecule.bonds_engine)

        # Check the exit status of the job; if failed restart the job up to 2 times
        restart_count = 1
        while not result['success'] and restart_count < 3:
            append_to_log(f'{molecule.bonds_engine} optimisation failed with error {result["error"]}; restarting',
                          msg_type='minor')

            # 1) If we have a file read error just start again
            if result['error'] == 'FileIO':
                result = qm_engine.generate_input(input_type='mm', optimise=True, restart=True, execute=molecule.bonds_engine)

            # 2) If we have a distance matrix error we should start from a different structure try the input
            elif result['error'] == 'Distance matrix' and restart_count == 1:
                result = qm_engine.generate_input(input_type='input', optimise=True, execute=molecule.bonds_engine)

            # 3) If we have already tried the starting structure generate a conformer and try again
            elif result['error'] == 'Distance matrix':
                molecule.write_pdb()
                molecule.coords['temp'] = RDKit().generate_conformers(f'{molecule.name}.pdb')[0]
                result = qm_engine.generate_input(input_type='temp', optimise=True, execute=molecule.bonds_engine)

            # Sometimes the user has not allowed enough iterations so try again
            elif result['error'] == 'Max iterations':
                result = qm_engine.generate_input(input_type='input', optimise=True, restart=True, execute=molecule.bonds_engine)

            restart_count += 1

        if not result['success']:
            raise OptimisationFailed(f"{molecule.bonds_engine} "
                                     f"optimisation did not converge after 3 restarts; last error {result['error']}")

        molecule.coords['qm'], molecule.qm_energy = qm_engine.optimised_structure()
        molecule.write_xyz(input_type='qm', name='opt')

        append_to_log(f'Finishing qm_optimisation of molecule{" using geometric" if molecule.geometric else ""}')

        return molecule

    def hessian(self, molecule):
        """Using the assigned bonds engine, calculate and extract the Hessian matrix."""

        append_to_log('Starting hessian calculation')
        molecule.find_bond_lengths(input_type='qm')

        # Check what engine to use
        if molecule.bonds_engine == 'g09' or molecule.bonds_engine == 'g16':
            qm_engine = self.engine_dict[molecule.bonds_engine](molecule)

            # Use the checkpoint file as this has higher xyz precision
            try:
                copy(os.path.join(molecule.home, os.path.join('03_qm_optimise', 'lig.chk')), 'lig.chk')
                result = qm_engine.generate_input(input_type='qm', hessian=True, restart=True, execute=molecule.bonds_engine)
            except FileNotFoundError:
                append_to_log('qm_optimise checkpoint not found, optimising first to refine atomic coordinates')
                result = qm_engine.generate_input(input_type='qm', optimise=True, hessian=True, execute=molecule.bonds_engine)

            if result['success']:
                molecule.hessian = qm_engine.hessian()
            else:
                raise HessianCalculationFailed('The hessian was not calculated check the log file.')

        else:
            qceng = QCEngine(molecule)
            hessian = qceng.call_qcengine('psi4', 'hessian', input_type='qm')

            np.savetxt('hessian.txt', hessian)

            molecule.hessian = hessian

        append_to_log(f'Finishing Hessian calculation using {molecule.bonds_engine}')

        return molecule

    @staticmethod
    def mod_sem(molecule):
        """Modified Seminario for bonds and angles."""

        append_to_log('Starting mod_Seminario method')

        mod_sem = ModSeminario(molecule)
        mod_sem.modified_seminario_method()

        molecule.symmetrise_bonded_parameters()

        append_to_log('Finishing Mod_Seminario method')

        return molecule

    def density(self, molecule):
        """Perform density calculation with the qm engine."""

        append_to_log('Starting density calculation')

        if molecule.density_engine == 'onetep':
            molecule.write_xyz(input_type='qm')
            # If using ONETEP, stop after this step
            append_to_log('Density analysis file made for ONETEP')

            # Edit the order to end here
            self.order = OrderedDict([('density', self.density), ('charges', self.skip), ('lennard_jones', self.skip),
                                      ('torsion_scan', self.skip), ('pause', self.pause),
                                      ('finalise', self.finalise)])

        else:
            qm_engine = self.engine_dict[molecule.density_engine](molecule)
            qm_engine.generate_input(input_type='qm', density=True, execute=molecule.density_engine)
            append_to_log('Finishing Density calculation')

        return molecule

    @staticmethod
    def charges(molecule):
        """Perform DDEC calculation with Chargemol."""

        append_to_log('Starting charge partitioning')
        copy(os.path.join(molecule.home, os.path.join('06_density', f'{molecule.name}.wfx')), f'{molecule.name}.wfx')
        c_mol = Chargemol(molecule)
        c_mol.generate_input()

        append_to_log(f'Finishing Charge partitioning with Chargemol and DDEC{molecule.ddec_version}')

        return molecule

    @staticmethod
    def lennard_jones(molecule):
        """Calculate Lennard-Jones parameters, and extract virtual sites."""

        append_to_log('Starting Lennard-Jones parameter calculation')

        charges_folder = os.path.join(molecule.home, '07_charges')
        for file in os.listdir(charges_folder):
            if file.startswith('DDEC'):
                copy(os.path.join(charges_folder, file), file)

        lj = LennardJones(molecule)
        molecule.NonbondedForce = lj.calculate_non_bonded_force()

        # This also now implies the opls combination rule
        molecule.combination = 'opls'

        append_to_log('Finishing Lennard-Jones parameter calculation')

        return molecule

    @staticmethod
    def torsion_scan(molecule):
        """Perform torsion scan."""
        append_to_log('Starting torsion_scans')

        scan = TorsionScan(molecule)
        # Check that we have a scan order for the molecule this should of been captured from the dihedral file
        scan.find_scan_order()

        scan.scan()

        append_to_log('Finishing torsion_scans')

        return molecule

    @staticmethod
    def torsion_optimise(molecule):
        """Perform torsion optimisation."""

        append_to_log('Starting torsion_optimisations')

        # First we should make sure we have collected the results of the scans
        if molecule.qm_scans is None:
            os.chdir(os.path.join(molecule.home, '09_torsion_scan'))
            scan = TorsionScan(molecule)
            if molecule.scan_order is None:
                scan.find_scan_order()
            scan.collect_scan()
            os.chdir(os.path.join(molecule.home, '10_torsion_optimise'))

        opt = TorsionOptimiser(molecule)
        opt.run()

        append_to_log('Finishing torsion_optimisations')

        return molecule

    @staticmethod
    def finalise(molecule):
        """
        Make the xml and pdb file;
        print the ligand object to terminal (in abbreviated form) and to the log file (unabbreviated).
        """

        molecule.write_pdb()
        molecule.write_parameters()

        pretty_print(molecule, to_file=True)
        pretty_print(molecule)

        return molecule

    @staticmethod
    def skip(molecule):
        """A blank method that does nothing to that stage but adds the pickle points to not break the flow."""

        return molecule

    @staticmethod
    def pause():
        """
        Pause the analysis when using ONETEP so we can come back into the workflow
        without breaking the pickling process.
        """

        printf('QUBEKit stopping at ONETEP step!\n To continue please move the ddec.onetep file and xyz file to the '
               'density folder and use QUBEKit -restart lennard_jones.')

        return

    @staticmethod
    def store_torsions(molecule, torsions_list=None):
        """
        Take the molecule object and the list of torsions and convert them to rotatable centres. Then, put them in the
        scan order object.
        """

        if torsions_list is not None:

            scan_order = []
            for torsion in torsions_list:
                tor = tuple(atom for atom in torsion.split('-'))
                # convert the string names to the index
                core = (molecule.get_atom_with_name(tor[1]).atom_index, molecule.get_atom_with_name(tor[2]).atom_index)

                if core in molecule.rotatable:
                    scan_order.append(core)
                elif reversed(core) in molecule.rotatable:
                    scan_order.append(reversed(core))

            molecule.scan_order = scan_order

        return molecule

    @staticmethod
    def torsion_test(molecule):
        """Take the molecule and do the torsion test method."""

        # If there is a constraints file we should move it
        if molecule.constraints_file is not None:
            copy(molecule.constraints_file, molecule.constraints_file.name)

        # Now we should run the torsion_test method
        opt = TorsionOptimiser(molecule)
        opt.torsion_test()

        printf('Torsion testing done!')

        # Now we must remove the torsion_test attrib
        del molecule.torsion_test

        return molecule


def main():
    """Putting class call inside a function squashes the __repr__ call"""
    ArgsAndConfigs()


if __name__ == '__main__':
    # For running with debugger, normal entry point is defined in setup.py
    main()



#!/usr/bin/env python3

from QUBEKit.utils import constants
from collections import OrderedDict
from copy import deepcopy

import xml.etree.ElementTree as ET


class Parametrisation:
    """
    Class of methods which perform the initial parametrisation for the molecule.
    The Parameters will be stored into the molecule as dictionaries as this is easy to manipulate and convert
    to a parameter tree.

    Note all parameters gathered here are indexed from 0,
    whereas the ligand object indices start from 1 for all networkx related properties such as bonds!


    Parameters
    ---------
    molecule : QUBEKit molecule object

    input_file : an OpenMM style xml file associated with the molecule object

    fftype : the FF type the molecule will be parametrised with
             only needed in the case of gaff or gaff2 else will be assigned based on class used.

    Returns
    -------
    AtomTypes : dictionary of the atom names, the associated OPLS type and class type stored under number.
                {0: [C00, OPLS_800, C800]}

    Residues : dictionary of residue names indexed by the order they appear.

    HarmonicBondForce : dictionary of equilibrium distances and force constants stored under the bond tuple.
                        {(0, 1): [eqr=456, fc=984375]}

    HarmonicAngleForce : dictionary of equilibrium  angles and force constant stored under the angle tuple.

    PeriodicTorsionForce : dictionary of periodicity, barrier and phase stored under the torsion tuple.

    NonbondedForce : dictionary of charge, sigma and epsilon stored under the original atom ordering.
    """

    def __init__(self, molecule, input_file=None, fftype=None):

        self.molecule = molecule
        self.input_file = input_file
        self.fftype = fftype
        self.molecule.combination = 'opls'
        self.combination = 'amber'

        # could be a problem for boron compounds
        self.molecule.AtomTypes = {}
        self.molecule.HarmonicBondForce = {bond: [0, 0] for bond in self.molecule.bond_lengths.keys()}
        self.molecule.HarmonicAngleForce = {angle: [0, 0] for angle in self.molecule.angle_values.keys()}
        self.molecule.NonbondedForce = OrderedDict((number, [0, 0, 0]) for number in range(len(self.molecule.atoms)))
        self.molecule.PeriodicTorsionForce = OrderedDict()

    def __repr__(self):
        return f'{self.__class__.__name__}({self.__dict__!r})'

    def gather_parameters(self):
        """
        This method parses the serialised xml file and collects the parameters ready to pass them
        to build tree.
        """

        # Try to gather the AtomTypes first
        for atom in self.molecule.atoms:
            self.molecule.AtomTypes[atom.atom_index] = [atom.atom_name, 'QUBE_' + str(000 + atom.atom_index),
                                                        str(atom.atomic_symbol) + str(000 + atom.atom_index)]

        try:
            in_root = ET.parse('serialised.xml').getroot()

            # Extract all bond data
            for Bond in in_root.iter('Bond'):
                bond = (int(Bond.get('p1')), int(Bond.get('p2')))
                if bond in self.molecule.HarmonicBondForce:
                    self.molecule.HarmonicBondForce[bond] = [float(Bond.get('d')), float(Bond.get('k'))]
                else:
                    self.molecule.HarmonicBondForce[bond[::-1]] = [float(Bond.get('d')), float(Bond.get('k'))]

            # Extract all angle data
            for Angle in in_root.iter('Angle'):
                angle = int(Angle.get('p1')), int(Angle.get('p2')), int(Angle.get('p3'))
                if angle in self.molecule.HarmonicAngleForce:
                    self.molecule.HarmonicAngleForce[angle] = [float(Angle.get('a')), float(Angle.get('k'))]
                else:
                    self.molecule.HarmonicAngleForce[angle[::-1]] = [float(Angle.get('a')), float(Angle.get('k'))]

            # Extract all non-bonded data
            i = 0
            for Atom in in_root.iter('Particle'):
                if "eps" in Atom.attrib:
                    self.molecule.NonbondedForce[i] = [float(Atom.get('q')), float(Atom.get('sig')), float(Atom.get('eps'))]
                    i += 1

            # Extract all of the torsion data
            phases = [0, constants.PI, 0, constants.PI]
            for Torsion in in_root.iter('Torsion'):
                tor_str_forward = tuple(int(Torsion.get(f'p{i}')) for i in range(1, 5))
                tor_str_back = tuple(reversed(tor_str_forward))

                if tor_str_forward not in self.molecule.PeriodicTorsionForce and tor_str_back not in self.molecule.PeriodicTorsionForce:
                    self.molecule.PeriodicTorsionForce[tor_str_forward] = [
                        [int(Torsion.get('periodicity')), float(Torsion.get('k')), phases[int(Torsion.get('periodicity')) - 1]]]

                elif tor_str_forward in self.molecule.PeriodicTorsionForce:
                    self.molecule.PeriodicTorsionForce[tor_str_forward].append(
                        [int(Torsion.get('periodicity')), float(Torsion.get('k')), phases[int(Torsion.get('periodicity')) - 1]])

                elif tor_str_back in self.molecule.PeriodicTorsionForce:
                    self.molecule.PeriodicTorsionForce[tor_str_back].append(
                        [int(Torsion.get('periodicity')), float(Torsion.get('k')), phases[int(Torsion.get('periodicity')) - 1]])

        except FileNotFoundError:
            # Check what parameter engine we are using if not none then raise an error
            if self.molecule.parameter_engine != 'none':
                raise FileNotFoundError('Molecule could not be serialised from openMM')
        # Now we have all of the torsions from the openMM system
        # we should check if any torsions we found in the molecule do not have parameters
        # if they don't give them the default 0 parameter this will not change the energy
        if self.molecule.dihedrals is not None:
            for tor_list in self.molecule.dihedrals.values():
                for torsion in tor_list:
                    if torsion not in self.molecule.PeriodicTorsionForce and tuple(reversed(torsion)) not in self.molecule.PeriodicTorsionForce:
                        self.molecule.PeriodicTorsionForce[torsion] = [[1, 0, 0], [2, 0, constants.PI],
                                                                       [3, 0, 0], [4, 0, constants.PI]]
        torsions = [sorted(key) for key in self.molecule.PeriodicTorsionForce.keys()]
        if self.molecule.improper_torsions is not None:
            for torsion in self.molecule.improper_torsions:
                if sorted(torsion) not in torsions:
                    # The improper torsion is missing and should be added with no energy
                    self.molecule.PeriodicTorsionForce[torsion] = [[1, 0, 0], [2, 0, constants.PI],
                                                                   [3, 0, 0], [4, 0, constants.PI]]

        # Now we need to fill in all blank phases of the Torsions
        for key, val in self.molecule.PeriodicTorsionForce.items():
            vns = [1, 2, 3, 4]
            if len(val) < 4:
                # now need to add the missing terms from the torsion force
                for force in val:
                    vns.remove(force[0])
                for i in vns:
                    val.append([i, 0, phases[int(i) - 1]])
        # sort by periodicity using lambda function
        for val in self.molecule.PeriodicTorsionForce.values():
            val.sort(key=lambda x: x[0])

        # now we need to tag the proper and improper torsions and reorder them so the first atom is the central
        improper_torsions = None
        if self.molecule.improper_torsions is not None:
            improper_torsions = OrderedDict()
            for improper in self.molecule.improper_torsions:
                for key, val in self.molecule.PeriodicTorsionForce.items():
                    # for each improper find the corresponding torsion parameters and save
                    if sorted(key) == sorted(improper):
                        # if they match tag the dihedral
                        self.molecule.PeriodicTorsionForce[key].append('Improper')
                        # replace the key with the strict improper order first atom is center
                        improper_torsions[improper] = val

        torsions = deepcopy(self.molecule.PeriodicTorsionForce)
        # Remake the torsion; store in the ligand
        self.molecule.PeriodicTorsionForce = OrderedDict((v, k) for v, k in torsions.items() if k[-1] != 'Improper')

        # Add the improper at the end of the torsion object
        if improper_torsions is not None:
            for key, val in improper_torsions.items():
                self.molecule.PeriodicTorsionForce[key] = val



#!/usr/bin/env python3

from .babel import Babel
from .chargemol import Chargemol
from QUBEKit.utils.exceptions import try_load


# Ensures that errors with particular engines are only raised if said engine is being used.
# No point raising "cannot execute psi4 command" if user is using g09. See try_load docstring for info.
Gaussian = try_load('Gaussian', 'QUBEKit.engines.gaussian')
ONETEP = try_load('ONETEP', 'QUBEKit.engines.onetep')
OpenMM = try_load('OpenMM', 'QUBEKit.engines.openmm')
PSI4 = try_load('PSI4', 'QUBEKit.engines.psi4')
QCEngine = try_load('QCEngine', 'QUBEKit.engines.qcengine')
RDKit = try_load('RDKit', 'QUBEKit.engines.rdkit')
Element = try_load('Element', 'QUBEKit.engines.rdkit')



#!/usr/bin/env python3

from QUBEKit.engines import RDKit
from QUBEKit.parametrisation.base_parametrisation import Parametrisation
from QUBEKit.utils.decorators import for_all_methods, timer_logger

from openforcefield.topology import Molecule, Topology
from openforcefield.typing.engines.smirnoff import ForceField

from simtk.openmm import app, XmlSerializer


@for_all_methods(timer_logger)
class OpenFF(Parametrisation):
    """
    This class uses the openFFtoolkit 2 to parametrise a molecule and load an OpenMM simulation.
    A serialised XML is then stored in the parameter dictionaries.
    """

    def __init__(self, molecule, input_file=None, fftype='frost'):
        super().__init__(molecule, input_file, fftype)

        self.serialise_system()
        self.gather_parameters()
        self.molecule.parameter_engine = 'OpenFF ' + self.fftype
        self.molecule.combination = self.combination

    def serialise_system(self):
        """Create the OpenMM system; parametrise using frost; serialise the system."""

        # Create an openFF molecule from the rdkit molecule
        try:
            off_molecule = Molecule.from_rdkit(self.molecule.rdkit_mol, allow_undefined_stereo=True)
        except AttributeError:
            raise AttributeError('An rdkit molecule object is required but could not be generated from the input file.')

        # Make the openMM system
        off_topology = off_molecule.to_topology()

        # Load the smirnoff99Frosst force field.
        forcefield = ForceField('test_forcefields/smirnoff99Frosst.offxml')

        # Parametrize the topology and create an OpenMM System.
        system = forcefield.create_openmm_system(off_topology)

        # Serialise the OpenMM system into the xml file
        with open('serialised.xml', 'w+') as out:
            out.write(XmlSerializer.serializeSystem(system))



#!/usr/bin/env python3

from QUBEKit.utils.helpers import pretty_print, unpickle

from datetime import datetime
from functools import wraps, partial
import logging
import os
from time import time


def timer_func(orig_func):
    """
    Prints the runtime of a function when applied as a decorator (@timer_func).
    Currently only used for debugging.
    """

    @wraps(orig_func)
    def wrapper(*args, **kwargs):

        t1 = time()
        result = orig_func(*args, **kwargs)
        t2 = time() - t1

        print(f'{orig_func.__qualname__} ran in: {t2} seconds.')

        return result
    return wrapper


def timer_logger(orig_func):
    """
    Logs the various timings of a function in a dated and numbered file.
    Writes the start time, function / method qualname and docstring when function / method starts.
    Then outputs the runtime and time when function / method finishes.
    """

    # TODO Currently only logs when the decorated class is passed molecule.
    #  Need a better way of getting log file beyond just getting it from the molecule object.

    @wraps(orig_func)
    def wrapper(*args, **kwargs):

        start_time = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        t1 = time()

        # All this does is check if the decorated function has a molecule attribute;
        # if it does, the function must therefore have a home path which can be used to write to the log file.
        if len(args) >= 1 and hasattr(args[0], 'molecule'):
            if getattr(args[0].molecule, 'home') is not None:
                log_file_path = os.path.join(args[0].molecule.home, 'QUBEKit_log.txt')
            else:
                return orig_func(*args, **kwargs)
        else:
            if os.path.exists('../QUBEKit_log.txt'):
                log_file_path = '../QUBEKit_log.txt'
            else:
                return orig_func(*args, **kwargs)

        with open(log_file_path, 'a+') as log_file:
            log_file.write(f'{orig_func.__qualname__} began at {start_time}.\n\n    ')
            log_file.write(f'Docstring for {orig_func.__qualname__}:\n    {orig_func.__doc__}\n\n')

            time_taken = time() - t1

            mins, secs = divmod(time_taken, 60)
            hours, mins = divmod(mins, 60)

            secs, remain = str(float(secs)).split('.')

            time_taken = f'{int(hours):02d}h:{int(mins):02d}m:{int(secs):02d}s.{remain[:5]}'
            end_time = datetime.now().strftime('%Y-%m-%d %H:%M:%S')

            log_file.write(f'{orig_func.__qualname__} finished in {time_taken} at {end_time}.\n\n')
            # Add some separation space between function / method logs.
            log_file.write(f'{"-" * 50}\n\n')

        return orig_func(*args, **kwargs)
    return wrapper


def for_all_methods(decorator):
    """
    Applies a decorator to all methods of a class (includes sub-classes and init; it is literally all callables).
    This class decorator is applied using '@for_all_methods(timer_func)' for example.
    """

    @wraps(decorator)
    def decorate(cls):
        # Examine all class attributes.
        for attr in cls.__dict__:
            # Check if each class attribute is a callable method.
            if callable(getattr(cls, attr)):
                # Set the callables to be decorated.
                setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls
    return decorate


def logger_format(log_file):
    """
    Creates logging object to be returned.
    Contains proper formatting and locations for logging exceptions.
    This isn't a decorator itself but is only used by exception_logger so it makes sense for it to be here.
    """

    logger = logging.getLogger('Exception Logger')
    logger.setLevel(logging.INFO)

    file_handler = logging.FileHandler(log_file)

    # Format the log message
    fmt = '\n\n%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    formatter = logging.Formatter(fmt)
    file_handler.setFormatter(formatter)

    logger.addHandler(file_handler)

    return logger


def exception_logger(func):
    """
    Decorator which logs exceptions to QUBEKit_log.txt file if one occurs.
    Do not apply this decorator to a function / method unless a log file will exist in the working dir;
    doing so will just raise the exception as normal.

    On any Exception, the Ligand class objects which are taken from the pickle file are printed to the log file,
    then the full stack trace is printed to the log file as well.

    Currently, only Execute.run is decorated like this, as it will always have a log file.
    Decorating other functions this way is possible and won't break anything, but it is pointless.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        # Any BaseException is logged; KeyboardInterrupt and SystemExit must still be raised (see below)
        except BaseException as exc:

            home = getattr(args[0].molecule, 'home', None)
            state = getattr(args[0].molecule, 'state', None)

            if home is None or state is None:
                raise

            mol = unpickle()[state]
            pretty_print(mol, to_file=True, finished=False)

            log_file = os.path.join(home, 'QUBEKit_log.txt')
            logger = logger_format(log_file)

            logger.exception(f'\nAn exception occurred with: {func.__qualname__}\n')
            print(f'\n\nAn exception occurred with: {func.__qualname__}\n'
                  f'Exception: {exc}\nView the log file for details.'.upper())

            if isinstance(exc, SystemExit) or isinstance(exc, KeyboardInterrupt):
                raise

            # Re-raises the exception if it's not a bulk run.
            # Even if the exception is not raised, it is still logged.
            if len(args) >= 1 and hasattr(args[0], 'molecule'):
                if not hasattr(args[0].molecule, 'bulk_run'):
                    raise
                if args[0].molecule.bulk_run is None:
                    raise

    return wrapper


class ExceptionLogger:
    """Alternate implementation of exception logger functions above"""

    def __init__(self, func):

        self.func = func
        self.log_file = None

    def __get__(self, instance, owner):
        """Allows decoration of functions and methods"""
        return partial(self.__call__, instance)

    def __call__(self, *args, **kwargs):

        try:
            return self.func(*args, **kwargs)
        # Any BaseException is logged; KeyboardInterrupt and SystemExit must still be raised (see below)
        except Exception as exc:

            home = getattr(args[0].molecule, 'home', None)
            state = getattr(args[0].molecule, 'state', None)

            if home is None or state is None:
                raise

            mol = unpickle()[state]
            pretty_print(mol, to_file=True, finished=False)

            self.log_file = os.path.join(home, 'QUBEKit_log.txt')
            logger = self.logger_format()

            logger.exception(f'\nAn exception occurred with: {self.func.__qualname__}\n')
            print(f'\n\nAn exception occurred with: {self.func.__qualname__}\n'
                  f'Exception: {exc}\nView the log file for details.'.upper())

            # Re-raises the exception if it's not a bulk run.
            # Even if the exception is not raised, it is still logged.
            if len(args) >= 1 and hasattr(args[0], 'molecule'):
                if not hasattr(args[0].molecule, 'bulk_run'):
                    raise
                if args[0].molecule.bulk_run is None:
                    raise

    def logger_format(self):

        logger = logging.getLogger('Exception Logger')
        logger.setLevel(logging.INFO)

        file_handler = logging.FileHandler(self.log_file)

        # Format the log message
        fmt = '\n\n%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        formatter = logging.Formatter(fmt)
        file_handler.setFormatter(formatter)

        logger.addHandler(file_handler)

        return logger



#!/usr/bin/env python3

from QUBEKit.utils.decorators import for_all_methods, timer_logger

import numpy as np

from rdkit import Chem
from rdkit.Chem import AllChem, Descriptors
from rdkit.Chem.rdchem import GetPeriodicTable
from rdkit.Chem.rdForceFieldHelpers import MMFFOptimizeMolecule, UFFOptimizeMolecule


@for_all_methods(timer_logger)
class RDKit:
    """Class for controlling useful RDKit functions."""
    def __init__(self):
        pass

    def read_file(self, filename):

        # Try and read the file
        if filename.suffix == '.pdb':
            mol = Chem.MolFromPDBFile(filename.name, removeHs=False)
        elif filename.suffix == '.mol2':
            mol = Chem.MolFromMol2File(filename.name, removeHs=False)
        elif filename.suffix == '.mol':
            mol = Chem.MolFromMolFile(filename.name, removeHs=False)
        else:
            mol = None

        return mol

    def smiles_to_rdkit_mol(self, smiles_string, name=None):
        """
        Converts smiles strings to RDKit mol object.
        :param smiles_string: The hydrogen free smiles string
        :param name: The name of the molecule this will be used when writing the pdb file
        :return: The RDKit molecule
        """

        mol = AllChem.MolFromSmiles(smiles_string)
        if name is None:
            name = input('Please enter a name for the molecule:\n>')
        mol.SetProp('_Name', name)
        mol_hydrogens = AllChem.AddHs(mol)
        AllChem.EmbedMolecule(mol_hydrogens, AllChem.ETKDG())
        AllChem.SanitizeMol(mol_hydrogens)

        return mol_hydrogens

    def mm_optimise(self, filename, ff='MMF'):
        """
        Perform rough preliminary optimisation to speed up later optimisations.
        :param filename: The name of the input file
        :param ff: The Force field to be used either MMF or UFF
        :return: The name of the optimised pdb file that is made
        """

        # Get the rdkit molecule
        mol = RDKit().read_file(filename)

        {'MMF': MMFFOptimizeMolecule, 'UFF': UFFOptimizeMolecule}[ff](mol)

        AllChem.MolToPDBFile(mol, f'{filename.stem}_rdkit_optimised.pdb')

        return f'{filename.stem}_rdkit_optimised.pdb'

    def rdkit_descriptors(self, rdkit_mol):
        """
        Use RDKit Descriptors to extract properties and store in Descriptors dictionary.
        :param rdkit_mol: The molecule input file
        :return: descriptors dictionary
        """

        # Use RDKit Descriptors to extract properties and store in Descriptors dictionary
        return {'Heavy atoms': Descriptors.HeavyAtomCount(rdkit_mol),
                'H-bond donors': Descriptors.NumHDonors(rdkit_mol),
                'H-bond acceptors': Descriptors.NumHAcceptors(rdkit_mol),
                'Molecular weight': Descriptors.MolWt(rdkit_mol),
                'LogP': Descriptors.MolLogP(rdkit_mol)}

    def get_smiles(self, filename):
        """
        Use RDKit to load in the pdb file of the molecule and get the smiles code.
        :param filename: The molecule input file
        :return: The smiles string
        """

        mol = RDKit().read_file(filename)

        return Chem.MolToSmiles(mol, isomericSmiles=True, allHsExplicit=True)

    def get_smarts(self, filename):
        """
        Use RDKit to get the smarts string of the molecule.
        :param filename: The molecule input file
        :return: The smarts string
        """

        mol = RDKit().read_file(filename)

        return Chem.MolToSmarts(mol)

    def get_mol(self, filename):
        """
        Use RDKit to generate a mol file.
        :param filename: The molecule input file
        :return: The name of the mol file made
        """

        mol = RDKit().read_file(filename)

        mol_name = f'{filename.stem}.mol'
        Chem.MolToMolFile(mol, mol_name)

        return mol_name

    def generate_conformers(self, filename, conformer_no=10):
        """
        Generate a set of x conformers of the molecule
        :param conformer_no: The amount of conformers made for the molecule
        :param filename: The name of the input file
        :return: A list of conformer position arrays
        """

        mol = RDKit().read_file(filename)

        cons = AllChem.EmbedMultipleConfs(mol, numConfs=conformer_no)
        positions = cons.GetConformers()

        return [conformer.GetPositions() for conformer in positions]

    def find_symmetry_classes(self, mol):
        """
        Generate list of tuples of symmetry-equivalent (homotopic) atoms in the molecular graph
        based on: https://sourceforge.net/p/rdkit/mailman/message/27897393/
        Our thanks to Dr Michal Krompiec for the symmetrisation method and its implementation.
        :param mol: molecule to find symmetry classes for (rdkit mol class object)
        :return: A dict where the keys are the atom indices and the values are their type
        (type is arbitrarily based on index; only consistency is needed, no specific values)
        """

        # Check CIPRank is present for first atom (can assume it is present for all afterwards)
        if not mol.GetAtomWithIdx(0).HasProp('_CIPRank'):
            Chem.AssignStereochemistry(mol, cleanIt=True, force=True, flagPossibleStereoCenters=True)

        # Array of ranks showing matching atoms
        cip_ranks = np.array([int(atom.GetProp('_CIPRank')) for atom in mol.GetAtoms()])

        # Map the ranks to the atoms to produce a list of symmetrical atoms
        atom_symmetry_classes = [np.where(cip_ranks == rank)[0].tolist() for rank in range(max(cip_ranks) + 1)]

        # Convert from list of classes to dict where each key is an atom and each value is its class (just a str)
        atom_symmetry_classes_dict = {}
        # i will be used to define the class (just index based)
        for i, sym_class in enumerate(atom_symmetry_classes):
            for atom in sym_class:
                atom_symmetry_classes_dict[atom] = str(i)

        return atom_symmetry_classes_dict

    # def get_rmsd(self, mol, coordinates):
    #     """
    #     Get the rmsd between the current rdkit molecule and the coordinates provided
    #     :param mol: rdkit representation of the molecule
    #     :param coordinates: array of the coordinates to compare to
    #     :return: the rmsd value
    #     """
    #
    # def add_confomer(self, mol, coordinates):
    #     """"""

class Element:
    """
    Simple wrapper class for getting element info using RDKit.
    """

    pt = GetPeriodicTable()

    def mass(self, identifier):
        return self.pt.GetAtomicWeight(identifier)

    def number(self, identifier):
        return self.pt.GetAtomicNumber(identifier)

    def name(self, identifier):
        return self.pt.GetElementSymbol(identifier)



#!/usr/bin/env python3

from QUBEKit.utils import constants
from QUBEKit.utils.decorators import for_all_methods, timer_logger
from QUBEKit.utils.helpers import check_net_charge, set_net

from collections import OrderedDict
import os

import numpy as np


@for_all_methods(timer_logger)
class LennardJones:

    def __init__(self, molecule):

        self.molecule = molecule

        # self.ddec_data is the DDEC molecule data in the format:
        # ['atom number', 'atom type', 'x', 'y', 'z', 'charge', 'x dipole', 'y dipole', 'z dipole', 'vol']
        # It will be extended and tweaked by each core method of this class.

        self.ddec_data = []

        self.epsilon_conversion = constants.BOHR_TO_ANGS ** 6
        self.epsilon_conversion *= constants.HA_TO_KCAL_P_MOL
        self.epsilon_conversion *= constants.KCAL_TO_KJ

        self.sigma_conversion = constants.ANGS_TO_NM

        self.non_bonded_force = {}

    def extract_params_chargemol(self):
        """
        From Chargemol output files, extract the necessary parameters for calculation of L-J.
        Desired format:
        ['atom number', 'atom type', 'x', 'y', 'z', 'charge', 'x_dipole', 'y_dipole', 'z_dipole', 'vol']
        All vals are float except atom number (int) and atom type (str).
        """

        if self.molecule.ddec_version == 6:
            net_charge_file_name = 'DDEC6_even_tempered_net_atomic_charges.xyz'

        elif self.molecule.ddec_version == 3:
            net_charge_file_name = 'DDEC3_net_atomic_charges.xyz'

        else:
            raise ValueError('Unsupported DDEC version; please use version 3 or 6.')

        if not os.path.exists(net_charge_file_name):
            raise FileNotFoundError(
                '\nCannot find the DDEC output file.\nThis could be indicative of several issues.\n'
                'Please check Chargemol is installed in the correct location and that the configs'
                ' point to that location.')

        with open(net_charge_file_name, 'r+') as charge_file:
            lines = charge_file.readlines()

        # Find number of atoms
        atom_total = int(lines[0])

        for pos, row in enumerate(lines):
            # Data marker:
            if 'The following XYZ' in row:
                start_pos = pos + 2
                break
        else:
            raise EOFError(f'Cannot find charge data in {net_charge_file_name}.')

        # Append the atom number and type, coords, charge, dipoles:
        for line in lines[start_pos: start_pos + atom_total]:
            a_number, a_type, *data = line.split()
            self.ddec_data.append([int(a_number), a_type] + [float(datum) for datum in data])

        charges = [atom[5] for atom in self.ddec_data]
        check_net_charge(charges, ideal_net=self.molecule.charge)

        r_cubed_file_name = 'DDEC_atomic_Rcubed_moments.xyz'

        with open(r_cubed_file_name, 'r+') as vol_file:
            lines = vol_file.readlines()

        vols = [float(line.split()[-1]) for line in lines[2:atom_total + 2]]

        for pos, atom in enumerate(self.ddec_data):
            atom.append(vols[pos])

    def extract_params_onetep(self):
        """
        From ONETEP output files, extract the necessary parameters for calculation of L-J.
        Desired format:
        ['atom number', 'atom type', 'x', 'y', 'z', 'charge', 'vol']
        All vals are float except atom number (int) and atom type (str).
        """

        # We know this from the molecule object self.molecule try to get the info from there
        for atom in self.molecule.atoms:
            self.ddec_data.append([atom.atom_index + 1, atom.atomic_symbol] +
                                  [self.molecule.coords['input'][atom.atom_index][i] for i in range(3)])

        # TODO Just move the ddec.onetep file instead? Handle this in run file?
        #   At very least, should use abspath
        # Second file contains the rest (charges, dipoles and volumes):
        with open(f'{"" if os.path.exists("ddec.onetep") else "iter_1/"}ddec.onetep', 'r') as file:
            lines = file.readlines()

        charge_pos, vol_pos = False, False
        for pos, line in enumerate(lines):

            # Charges marker in file:
            if 'DDEC density' in line:
                charge_pos = pos + 7

            # Volumes marker in file:
            if 'DDEC Radial' in line:
                vol_pos = pos + 4

        if not (charge_pos and vol_pos):
            raise EOFError('Cannot locate charges and / or volumes in ddec.onetep file.')
            
        charges = [float(line.split()[-1]) for line in lines[charge_pos: charge_pos + len(self.ddec_data)]]
        check_net_charge(charges, ideal_net=self.molecule.charge)

        # Add the AIM-Valence and the AIM-Core to get V^AIM
        volumes = [float(line.split()[2]) + float(line.split()[3]) for line in lines[vol_pos: vol_pos + len(self.ddec_data)]]

        # Add the charges and volumes to the end of the inner lists (containing coords etc)
        for pos, atom in enumerate(self.ddec_data):
            atom.extend((charges[pos], volumes[pos]))

    def append_ais_bis(self):
        """
        Use the AIM parameters from extract_params_*() to calculate a_i and b_i according to paper.
        Calculations from paper have been combined and simplified for faster computation.
        """

        # Beware weird units, (wrong in the paper too).
        # 'elem' : [vfree, bfree, rfree]
        # Units: [vfree: Bohr ** 3, bfree: Ha * (Bohr ** 6), rfree: Angs]

        elem_dict = {
            'H': [7.6, 6.5, 1.64],
            'B': [46.7, 99.5, 2.08],
            'C': [34.4, 46.6, 2.08],
            'N': [25.9, 24.2, 1.72],
            'O': [22.1, 15.6, 1.60],
            'F': [18.2, 9.5, 1.58],
            'P': [84.6, 185, 2.00],
            'S': [75.2, 134.0, 2.00],
            'Cl': [65.1, 94.6, 1.88],
            'Br': [95.7, 162.0, 1.96],
        }

        for pos, atom in enumerate(self.ddec_data):
            try:
                # r_aim = r_free * ((vol / v_free) ** (1 / 3))
                r_aim = elem_dict[f'{atom[1]}'][2] * ((atom[-1] / elem_dict[f'{atom[1]}'][0]) ** (1 / 3))

                # b_i = bfree * ((vol / v_free) ** 2)
                b_i = elem_dict[f'{atom[1]}'][1] * ((atom[-1] / elem_dict[f'{atom[1]}'][0]) ** 2)

                a_i = 32 * b_i * (r_aim ** 6)

                self.ddec_data[pos] += [r_aim, b_i, a_i]

            except KeyError:
                self.ddec_data[pos] += [0, 0, 0]

    def calculate_sig_eps(self):
        """
        Adds the sigma, epsilon terms to the ligand class object as a dictionary.
        The ligand class object (NonbondedForce) is stored as an empty dictionary until this method is called.
        first_pass argument prevents the sigmas being recalculated (unlike the epsilons).
        """

        # Creates Nonbondedforce dict for later xml creation.
        # Format: {0: [charge, sigma, epsilon], 1: [charge, sigma, epsilon], ... }
        # This follows the usual ordering of the atoms such as in molecule.coords.

        for pos, atom in enumerate(self.ddec_data):

            if atom[-1]:
                # sigma = (a_i / b_i) ** (1 / 6)
                sigma = (atom[-1] / atom[-2]) ** (1 / 6)
                sigma *= self.sigma_conversion

                # epsilon = (b_i ** 2) / (4 * a_i)
                epsilon = (atom[-2] ** 2) / (4 * atom[-1])
                epsilon *= self.epsilon_conversion

            else:
                sigma = epsilon = 0

            self.non_bonded_force[pos] = [atom[5], sigma, epsilon]

    def correct_polar_hydrogens(self):
        """
        Identifies the polar Hydrogens and changes the a_i, b_i values accordingly.
        May be removed / heavily changed if we switch away from atom typing and use SMARTS.
        """

        # Loop through pairs in topology
        # Create new pair list with the atoms
        new_pairs = []
        for pair in self.molecule.topology.edges:
            new_pair = (self.molecule.atoms[pair[0]], self.molecule.atoms[pair[1]])
            new_pairs.append(new_pair)

        # Find all the polar hydrogens and store their positions / atom numbers
        polars = []
        for pair in new_pairs:
            if 'O' == pair[0].atomic_symbol or 'N' == pair[0].atomic_symbol or 'S' == pair[0].atomic_symbol:
                if 'H' == pair[1].atomic_symbol:
                    polars.append(pair)

            if 'O' == pair[1].atomic_symbol or 'N' == pair[1].atomic_symbol or 'S' == pair[1].atomic_symbol:
                if 'H' == pair[0].atomic_symbol:
                    polars.append(pair)

        # Find square root of all b_i values so that they can be added easily according to paper's formula.
        for atom in self.ddec_data:
            atom[-2] = (atom[-2]) ** 0.5

        if polars:
            for pair in polars:
                if 'H' == pair[0].atomic_symbol or 'H' == pair[1].atomic_symbol:
                    if 'H' == pair[0].atomic_symbol:
                        polar_h_pos = pair[0].atom_index
                        polar_son_pos = pair[1].atom_index
                    else:
                        polar_h_pos = pair[1].atom_index
                        polar_son_pos = pair[0].atom_index

                    # Calculate the new b_i for the two polar atoms (polar h and polar sulfur, oxygen or nitrogen)
                    self.ddec_data[polar_son_pos][-2] += self.ddec_data[polar_h_pos][-2]
                    self.ddec_data[polar_h_pos][-2] = 0

        # Square all the b_i values again
        for atom in self.ddec_data:
            atom[-2] *= atom[-2]

        # Recalculate the a_i values
        for atom in self.ddec_data:
            atom[-1] = 32 * atom[-2] * (atom[-3] ** 6)

        # Update epsilon (not sigma) according to new a_i and b_i values
        for pos, atom in enumerate(self.ddec_data):

            if atom[-1] == 0:
                epsilon, self.non_bonded_force[pos][1] = 0, 0
            else:
                # epsilon = (b_i ** 2) / (4 * a_i)
                epsilon = (atom[-2] ** 2) / (4 * atom[-1])
                epsilon *= self.epsilon_conversion

            self.non_bonded_force[pos] = [atom[5], self.non_bonded_force[pos][1], epsilon]

    def apply_symmetrisation(self):
        """
        Using the atoms picked out to be symmetrised,
        apply the symmetry to the charge, sigma and epsilon values.
        """

        atom_types = {}
        for key, val in self.molecule.atom_symmetry_classes.items():
            atom_types.setdefault(val, []).append(key)

        # get the values to be symmetrised
        for sym_set in atom_types.values():
            charges, sigmas, epsilons = [], [], []
            if len(sym_set) > 1:
                for atom in sym_set:
                    charges.append(self.non_bonded_force[atom][0])
                    sigmas.append(self.non_bonded_force[atom][1])
                    epsilons.append(self.non_bonded_force[atom][2])

                # calculate the average values to be used in symmetry
                charge, sigma, epsilon = sum(charges) / len(charges), sum(sigmas) / len(sigmas), sum(epsilons) / len(epsilons)

                # Loop through the atoms again and store the new values
                for atom in sym_set:
                    self.non_bonded_force[atom] = [round(charge, 6), round(sigma, 6), round(epsilon, 6)]

        # make sure the net charge is correct for the current precision
        charges = [non_bonded[0] for non_bonded in self.non_bonded_force.values()]
        new_charges = set_net(charges, self.molecule.charge, 6)
        # Put the new charges back into the holder
        for non_bonded in zip(self.non_bonded_force.values(), new_charges):
            non_bonded[0][0] = non_bonded[1]

    def extract_extra_sites(self):
        """
        1) Gather the extra sites from the XYZ find parent and 2 reference atoms
        2) calculate the local coords site
        3) save the charge
        4) return back to the molecule
        (users have the option to use sites or no sites this way)
        """

        # weighting arrays for the virtual sites should not be changed
        w1o, w2o, w3o = 1.0, 0.0, 0.0  # SUM SHOULD BE 1
        w1x, w2x, w3x = -1.0, 1.0, 0.0  # SUM SHOULD BE 0
        w1y, w2y, w3y = -1.0, 0.0, 1.0  # SUM SHOULD BE 0

        if not os.path.exists('xyz_with_extra_point_charges.xyz'):
            return

        with open('xyz_with_extra_point_charges.xyz') as xyz_sites:
            lines = xyz_sites.readlines()

        sites = OrderedDict()
        sites_no = 0

        for i, line in enumerate(lines[2:]):
            # get the current element
            element = str(line.split()[0])

            if element != 'X':
                # search the following entries for sites connected to this atom
                for pos_site in lines[i + 3:]:
                    if str(pos_site.split()[0]) != 'X':
                        # if there are no sites break and start the next loop
                        break
                    else:
                        # get the virtual site coords
                        v_pos = np.array([float(pos_site.split()[x]) for x in range(1, 4)])
                        # get parent index number for the topology network
                        parent = i - sites_no
                        # get the two closest atoms to the parent
                        closest_atoms = list(self.molecule.topology.neighbors(parent))
                        if len(closest_atoms) < 2:
                            # find another atom if we only have one
                            # dont want to get the parent as a close atom
                            closest_atoms.append(list(self.molecule.topology.neighbors(closest_atoms[0]))[-1])

                        # Get the xyz coordinates of the reference atoms
                        parent_pos = self.molecule.coords['qm'][parent]
                        close_a = self.molecule.coords['qm'][closest_atoms[0]]
                        close_b = self.molecule.coords['qm'][closest_atoms[1]]

                        # work out the local coordinates site using rules from the OpenMM guide
                        orig = w1o * parent_pos + w2o * close_a + close_b * w3o
                        ab = w1x * parent_pos + w2x * close_a + w3x * close_b  # rb-ra
                        ac = w1y * parent_pos + w2y * close_a + w3y * close_b  # rb-ra
                        # Get the axis unit vectors
                        z_dir = np.cross(ab, ac)
                        z_dir = z_dir / np.sqrt(np.dot(z_dir, z_dir.reshape(3, 1)))
                        x_dir = ab / np.sqrt(np.dot(ab, ab.reshape(3, 1)))
                        y_dir = np.cross(z_dir, x_dir)
                        # Get the local coordinates positions
                        p1 = np.dot((v_pos - orig), x_dir.reshape(3, 1))
                        p2 = np.dot((v_pos - orig), y_dir.reshape(3, 1))
                        p3 = np.dot((v_pos - orig), z_dir.reshape(3, 1))

                        charge = float(pos_site.split()[4])

                        # store the site info [(parent top no, a, b), (p1, p2, p3), charge]]
                        sites[sites_no] = [(parent, closest_atoms[0], closest_atoms[1]), (p1 / 10, p2 / 10, p3 / 10), charge]
                        sites_no += 1

        self.molecule.sites = sites

        # get the parent non bonded values
        for site in sites.values():
            charge, sigma, eps = self.non_bonded_force[site[0][0]]
            # Change the charge on the first entry
            charge -= site[2]
            self.non_bonded_force[site[0][0]] = [charge, sigma, eps]

    def calculate_non_bonded_force(self):
        """
        Main worker method for LennardJones class. Extracts necessary parameters from ONETEP or Chargemol files;
        Calculates the a_i and b_i values;
        Calculates the sigma and epsilon values using those a_i and b_i values;
        Redistributes L-J parameters according to polar Hydrogens, then recalculates epsilon values.
        returns non_bonded_force for the XML creator in Ligand class.
        """

        if self.molecule.charges_engine == 'chargemol':
            self.extract_params_chargemol()

        elif self.molecule.charges_engine == 'onetep':
            self.extract_params_onetep()

        else:
            raise KeyError('Invalid charges engine provided, cannot extract charges.')

        # Calculate initial a_is and b_is
        self.append_ais_bis()

        # Use the a_is and b_is to calculate the non_bonded_force dict
        self.calculate_sig_eps()

        # Tweak for polar Hydrogens
        self.correct_polar_hydrogens()

        # Tweak the charge, sigma and epsilon for symmetry
        self.apply_symmetrisation()

        # Find extra site positions in local coords if present and tweak the charges of the parent
        if self.molecule.charges_engine == 'onetep':
            self.extract_extra_sites()

        return self.non_bonded_force



#!/usr/bin/env python3

from QUBEKit.ligand import Protein
from QUBEKit.parametrisation import XMLProtein
from QUBEKit.lennard_jones import LennardJones
from QUBEKit.protein_tools import qube_general, pdb_reformat, get_water

import argparse
from functools import partial
import os
import sys


printf = partial(print, flush=True)


def main():
    """
    This script is used to prepare proteins with the QUBE FF
    1) prepare the protein for onetep using --setup which prints an xyz of the system
    2) after the onetep calculation bring back the ddec.onetep file and parametrise the system with --build
    this must be the same pdb used in setup as the atom order must be retained.
    """

    # Setup the action classes
    class SetupAction(argparse.Action):
        """This class is called when we setup a new protein."""

        def __call__(self, pars, namespace, values, option_string=None):
            """This function is executed when setup is called."""

            printf('starting protein prep, reading pdb file...')
            protein = Protein(values)
            printf(f'{len(protein.Residues)} residues found!')
            # TODO find the magic numbers for the box for onetep
            protein.write_xyz(name='protein')
            printf(f'protein.xyz file made for ONETEP\n Run this file')
            sys.exit()

    class BuildAction(argparse.Action):
        """This class handles the building of the protein xml and pdb files."""

        def __call__(self, pars, namespace, values, option_string=None):
            """This function is executed when build is called."""

            pro = Protein(values)
            # print the QUBE general FF to use in the parametrisation
            qube_general()
            # now we want to add the connections and parametrise the protein
            XMLProtein(pro)
            # this updates the bonded info that is now in the object

            # finally we need the non-bonded parameters from onetep
            # TODO should we also have the ability to get DDEC6 charges from the cube file?
            pro.charge = 0
            pro.charges_engine = 'onetep'
            pro.density_engine = 'onetep'
            lj = LennardJones(pro)
            pro.NonbondedForce = lj.calculate_non_bonded_force()

            # now we write out the final parameters
            # we should also calculate the charges and lj at this point!
            printf('Writing pdb file with connections...')
            pro.write_pdb(name='QUBE_pro')
            printf('Writing XML file for the system...')
            pro.write_parameters(name='QUBE_pro', is_protein=True)
            # now remove the qube general file
            os.remove('QUBE_general_pi.xml')
            printf('Done')
            sys.exit()

    class WaterAction(argparse.Action):
        """This class builds the water models requested"""

        def __call__(self, pars, namespace, values, option_string=None):
            """This function is executed when water is called."""
            get_water(values)
            sys.exit()

    class ConvertAction(argparse.Action):
        """This class converts the names in a qube taj file to match the reference."""

        def __call__(self, pars, namespace, values, option_string=None):
            """This function is executed when water is called."""
            reference, target = values
            printf(reference, target)
            printf(f'Rewriting input: {target} to match: {reference}...')
            pdb_reformat(reference, target)
            printf('Done output made: QUBE_traj.pdb')
            sys.exit()

    # Set up the argument parser
    parser = argparse.ArgumentParser(prog='QUBEKit-pro', formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description='QUBEKit-pro is an extension to QUBEKit that allows the preparation'
                                                 'of amber prepared proteins with the QUBE forcefield and creates force'
                                                 'field xml files for OpenMM simulation.')
    parser.add_argument('-setup', '--setup', action=SetupAction,
                        help='Enter the name of the amber prepared protein file to convert to xyz for onetep.')

    parser.add_argument('-build', '--build_simulation_files', action=BuildAction,
                        help='Enter the name of the amber prepared protein file to generate QUBE pdb and xml files.')

    parser.add_argument('-water', '--water', nargs='?', default='help', action=WaterAction,
                        choices=['help', 'tip3p', 'spce', 'tip4p', 'tip4pew', 'tip5p', 'tip3pfb', 'tip4pfb', 'tip4p-d',
                                 'opc'],
                        help='Enter the name of the water model you would like to use.')

    parser.add_argument('-convert', '--convert', nargs=2, action=ConvertAction,
                        help='Enter the reference followed by the qube traj file you want to re format.')

    parser.parse_args()


if __name__ == '__main__':
    main()



"""
Get the hessians from the VEHICLe dataset

Get the json dict from qcportal
From this, initialise a Ligand object, add in the qm-optimised coords and hessian
Generate a

"""

from QUBEKit.engines import RDKit
from QUBEKit.ligand import Ligand
from QUBEKit.mod_seminario import ModSeminario
from QUBEKit.parametrisation.base_parametrisation import Parametrisation
from QUBEKit.utils import constants

import os
from pathlib import Path
import pickle

import numpy as np
import qcportal as ptl

from rdkit import DataStructs
from rdkit.Chem import AllChem


class BondsAndAngles:

    def __init__(self):
        self.client, self.ds, self.records = None, None, None
        self.get_records()

        self.bond_data = []
        self.bond_labels = []

        self.angle_data = []
        self.angle_labels = []

    def get_records(self):
        if os.path.exists('records.pkl'):
            with open('records.pkl', 'rb') as rec_file:
                self.client, self.ds, self.records = pickle.load(rec_file)
        else:
            # Cache VEHICLe dataset
            self.client = ptl.FractalClient()
            self.client.list_collections('OptimizationDataset')
            self.ds = self.client.get_collection('OptimizationDataset', 'OpenFF VEHICLe Set 1')
            self.ds.df.head()
            self.ds.list_specifications()

            # Cache records
            self.records = self.ds.data.records
            pickle_data = (self.client, self.ds, self.records)

            with open('records.pkl', 'wb') as rec_file:
                pickle.dump(pickle_data, rec_file)

    def loop_over_mols(self):
        for i, item in enumerate(self.records):
            ptl_name = item.strip('\n')
            smiles = self.records[ptl_name].attributes['canonical_smiles']
            print(f'iter: {str(i).zfill(5)}; smiles: {smiles}')

            # Specific ID of given smiles string
            try:
                opt_record = self.ds.get_entry(ptl_name).object_map['default']
            except KeyError:
                continue

            # Optimisation of molecule at ID: opt_record
            optimisation = self.client.query_procedures(id=opt_record)[0]

            # Extract hessian
            try:
                hessian = self.client.query_results(molecule=optimisation.final_molecule, driver='hessian')[0].return_result
            except IndexError:
                # Molecule has been optimised but no hessian has been calculated yet
                continue

            # Reshape hessian
            conversion = constants.HA_TO_KCAL_P_MOL / (constants.BOHR_TO_ANGS ** 2)
            hessian = np.array(hessian).reshape(int(len(hessian) ** 0.5), -1) * conversion

            # Extract optimised structure
            opt_struct = self.client.query_procedures(id=opt_record)[0].get_final_molecule()

            mol = self.calc_mod_sem(smiles, hessian, opt_struct)
            finger_prints = self.calc_fingerprint()
            self.format_data(mol, finger_prints)
            # Pickle the results every iteration so that an error won't mean progress is lost.
            # dedent if this isn't wanted.
            self.pickle_results()

    @staticmethod
    def calc_mod_sem(smiles, hessian, opt_struct):

        # Initialise Ligand object using the json dict from qcengine
        mol = Ligand(opt_struct.json_dict(), name='initial_test')

        # Set the qm coords to the input coords from qcengine
        mol.coords['qm'] = mol.coords['input']

        # Insert hessian and optimised coordinates
        mol.hessian = hessian
        mol.parameter_engine = 'none'

        # Create empty parameter dicts
        Parametrisation(mol).gather_parameters()

        with open('Modified_Seminario_Bonds.txt', 'a+') as bonds_file,\
                open('Modified_Seminario_Angles.txt', 'a+') as angles_file:
            bonds_file.write(f'\n\n{smiles}\n\n')
            angles_file.write(f'\n\n{smiles}\n\n')

        # Get Mod Sem angle and bond params
        ModSeminario(mol).modified_seminario_method()

        # Write a pdb to preserve order for RDKit
        mol.write_pdb(name='test')

        return mol

    @staticmethod
    def calc_fingerprint():

        rdkit_mol = RDKit().read_file(Path('test.pdb'))

        finger_prints = {}
        for atom in rdkit_mol.GetAtoms():
            atom_index = atom.GetIdx()
            fp = AllChem.GetHashedAtomPairFingerprintAsBitVect(rdkit_mol, maxLength=4, fromAtoms=[atom_index])
            arr = np.zeros(1, )
            DataStructs.ConvertToNumpyArray(fp, arr)
            finger_prints[atom_index] = arr

        os.system('rm test.pdb')

        return finger_prints

    def format_data(self, mol, finger_prints):

        for bond, val in mol.HarmonicBondForce.items():
            self.bond_data.append(np.concatenate((finger_prints[bond[0]], finger_prints[bond[1]])))
            self.bond_labels.append(val)

        for angle, val in mol.HarmonicAngleForce.items():
            self.angle_data.append(np.concatenate((finger_prints[angle[0]], finger_prints[angle[1]], finger_prints[angle[2]])))
            self.angle_labels.append(val)

    def pickle_results(self):
        with open('bond_train.pkl', 'wb') as bond_data_file, open('bond_label.pkl', 'wb') as bond_label_file:
            pickle.dump(self.bond_data, bond_data_file)
            pickle.dump(self.bond_labels, bond_label_file)

        with open('angle_train.pkl', 'wb') as angle_data_file, open('angle_label.pkl', 'wb') as angle_label_file:
            pickle.dump(self.angle_data, angle_data_file)
            pickle.dump(self.angle_labels, angle_label_file)


# client = ptl.FractalClient()
# client.list_collections('OptimizationDataset')
# ds = client.get_collection('OptimizationDataset', 'OpenFF VEHICLe Set 1')
# ds.df.head()
# ds.list_specifications()
# records = ds.data.records
#
# bond_data = []
# bond_label = []
#
# angle_data = []
# angle_label = []
#
# for i, item in enumerate(records):
#     ptl_name = item.strip('\n')
#     smiles = records[ptl_name].attributes['canonical_smiles']
#     print(f'iter: {str(i).zfill(5)}; smiles: {smiles}')
#
#     # Specific ID of given smiles string
#     try:
#         opt_record = ds.get_entry(ptl_name).object_map['default']
#     except KeyError:
#         continue
#
#     # Optimisation of molecule at ID: opt_record
#     optimisation = client.query_procedures(id=opt_record)[0]
#
#     # Extract hessian
#     try:
#         hessian = client.query_results(molecule=optimisation.final_molecule, driver='hessian')[0].return_result
#     except IndexError:
#         # Molecule has been optimised but no hessian has been calculated yet
#         continue
#
#     # Reshape hessian
#     conversion = constants.HA_TO_KCAL_P_MOL / (constants.BOHR_TO_ANGS ** 2)
#     hessian = np.array(hessian).reshape(int(len(hessian) ** 0.5), -1) * conversion
#
#     # Extract optimised structure
#     opt_struct = client.query_procedures(id=opt_record)[0].get_final_molecule()
#
#     # Initialise Ligand object using the json dict from qcengine
#     mol = Ligand(opt_struct.json_dict(), name='initial_test')
#
#     # Set the qm coords to the input coords from qcengine
#     mol.coords['qm'] = mol.coords['input']
#
#     # Insert hessian and optimised coordinates
#     mol.hessian = hessian
#     mol.parameter_engine = 'none'
#
#     # Create empty parameter dicts
#     Parametrisation(mol).gather_parameters()
#
#     with open('Modified_Seminario_Bonds.txt', 'a+') as bonds_file, \
#             open('Modified_Seminario_Angles.txt', 'a+') as angles_file:
#         bonds_file.write(f'\n\n{smiles}\n\n')
#         angles_file.write(f'\n\n{smiles}\n\n')
#
#     # Get Mod Sem angle and bond params
#     ModSeminario(mol).modified_seminario_method()
#
#     mol.write_pdb(name='test')
#
#     rdkit_mol = RDKit().read_file(Path('test.pdb'))
#
#     finger_prints = {}
#     for atom in rdkit_mol.GetAtoms():
#         atom_index = atom.GetIdx()
#         fp = AllChem.GetHashedAtomPairFingerprintAsBitVect(rdkit_mol, maxLength=4, fromAtoms=[atom_index])
#         arr = np.zeros(1, )
#         DataStructs.ConvertToNumpyArray(fp, arr)
#         finger_prints[atom_index] = arr
#
#     for bond, val in mol.HarmonicBondForce.items():
#         bond_data.append(np.concatenate((finger_prints[bond[0]], finger_prints[bond[1]])))
#         bond_label.append(val)
#
#     for angle, val in mol.HarmonicAngleForce.items():
#         angle_data.append(np.concatenate((finger_prints[angle[0]], finger_prints[angle[1]], finger_prints[angle[2]])))
#         angle_label.append(val)
#
#     os.system('rm test.pdb')
#
# with open('bond_train.pkl', 'wb') as bond_data_file, open('bond_label.pkl', 'wb') as bond_label_file:
#     pickle.dump(bond_data, bond_data_file)
#     pickle.dump(bond_label, bond_label_file)
#
# with open('angle_train.pkl', 'wb') as angle_data_file, open('angle_label.pkl', 'wb') as angle_label_file:
#     pickle.dump(angle_data, angle_data_file)
#     pickle.dump(angle_label, angle_label_file)
#
# data = pickle.load('train_data.pkl')
# labels = pickle.load('train_label.pkl')

if __name__ == '__main__':
    BondsAndAngles().loop_over_mols()
